#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron.db import api as db_api
from neutron.plugins.common import constants as pconst
from neutron.quota import resource_registry
from oslo_config import cfg
from oslo_log import helpers as log
from oslo_log import log as logging
from oslo_utils import excutils

from gbpservice._i18n import _LE
from gbpservice._i18n import _LI
from gbpservice._i18n import _LW
from gbpservice.common import utils
from gbpservice.neutron.db import servicechain_db
from gbpservice.neutron.services.grouppolicy.common import constants as gp_cts
from gbpservice.neutron.services.grouppolicy.common import utils as gutils
from gbpservice.neutron.services.servicechain.plugins.ncp import (
    context as ctx)
from gbpservice.neutron.services.servicechain.plugins.ncp import (
    exceptions as exc)
from gbpservice.neutron.services.servicechain.plugins.ncp import (
    node_driver_manager as manager)
from gbpservice.neutron.services.servicechain.plugins import sharing

LOG = logging.getLogger(__name__)

PLUMBER_NAMESPACE = 'gbpservice.neutron.servicechain.ncp_plumbers'
cfg.CONF.import_opt('policy_drivers',
                    'gbpservice.neutron.services.grouppolicy.config',
                    group='group_policy')
STATUS = 'status'
STATUS_DETAILS = 'status_details'
STATUS_SET = set([STATUS, STATUS_DETAILS])


class NodeCompositionPlugin(servicechain_db.ServiceChainDbPlugin,
                            sharing.SharingMixin):

    """Implementation of the Service Chain Plugin.

    """
    supported_extension_aliases = ["servicechain"]
    path_prefix = gp_cts.GBP_PREFIXES[pconst.SERVICECHAIN]

    @resource_registry.tracked_resources(
        servicechain_node=servicechain_db.ServiceChainNode,
        servicechain_spec=servicechain_db.ServiceChainSpec,
        servicechain_instance=servicechain_db.ServiceChainInstance,
        service_profile=servicechain_db.ServiceProfile)
    def __init__(self):
        self.driver_manager = manager.NodeDriverManager()
        super(NodeCompositionPlugin, self).__init__()
        self.driver_manager.initialize()
        plumber_klass = cfg.CONF.node_composition_plugin.node_plumber
        self.plumber = utils.load_plugin(
            PLUMBER_NAMESPACE, plumber_klass)
        self.plumber.initialize()
        LOG.info(_LI("Initialized node plumber '%s'"), plumber_klass)

    @log.log_method_call
    def create_servicechain_instance(self, context, servicechain_instance):
        """Instance created.

        When a Servicechain Instance is created, all its nodes need to be
        instantiated.
        """
        instance = self._process_commit_phase(context)
        if instance:
            return instance

        session = context.session
        deployers = {}
        with session.begin(subtransactions=True):
            instance = super(NodeCompositionPlugin,
                             self).create_servicechain_instance(
                                 context, servicechain_instance)
            if len(instance['servicechain_specs']) > 1:
                raise exc.OneSpecPerInstanceAllowed()
            deployers = self._get_scheduled_drivers(context, instance,
                                                    'deploy')
        if not gutils.is_precommit_policy_driver_configured():
            # Actual node deploy
            try:
                self._deploy_servicechain_nodes(context, deployers)
            except Exception:
                # Some node could not be deployed
                with excutils.save_and_reraise_exception():
                    LOG.error(_LE("Node deployment failed, "
                                  "deleting servicechain_instance %s"),
                              instance['id'])
                    self.delete_servicechain_instance(context, instance['id'])

        return instance

    def _process_commit_phase(self, context):
        if hasattr(context, 'commit_phase'):
            if not gutils.is_precommit_policy_driver_configured() and (
                context.commit_phase == gp_cts.PRE_COMMIT):
                # The following is a bit of a hack to no-op
                # the call from the postcommit policy driver
                # during the pre-commit phase.
                return True
            if gutils.is_precommit_policy_driver_configured() and (
                context.commit_phase == gp_cts.POST_COMMIT):
                instance = self.get_servicechain_instance(
                    context, context.servicechain_instance['id'])
                self._call_deploy_sc_node(context, instance)
                return instance

    def _call_deploy_sc_node(self, context, instance):
        # Actual node deploy
        try:
            deployers = self._get_scheduled_drivers(
                context, instance, 'deploy')
            self._deploy_servicechain_nodes(context, deployers)
        except Exception:
            # Some node could not be deployed
            with excutils.save_and_reraise_exception():
                LOG.error(_LE("Node deployment failed, "
                              "servicechain_instance %s is in ERROR state"),
                          instance['id'])

    @log.log_method_call
    def get_servicechain_instance(self, context,
                                  servicechain_instance_id, fields=None):
        """Instance retrieved.

        While get Servicechain Instance details, get all nodes details.
        """
        return self._get_resource(context, 'servicechain_instance',
                                  servicechain_instance_id, fields)

    @log.log_method_call
    def get_servicechain_instance(self, context,
                                  servicechain_instance_id, fields=None):
        """Instance retrieved.

        While get Servicechain Instance details, get all nodes details.
        """
        return self._get_resource(context, 'servicechain_instance',
                                  servicechain_instance_id, fields)

    @log.log_method_call
    def update_servicechain_instance(self, context, servicechain_instance_id,
                                     servicechain_instance):
        """Instance updated.

        When a Servicechain Instance is updated and the spec changed, all the
        nodes of the previous spec should be destroyed and the newer ones
        created.
        """
        instance = self._process_commit_phase(context)
        if instance:
            return instance

        session = context.session
        deployers = {}
        updaters = {}
        destroyers = {}
        with session.begin(subtransactions=True):
            original_instance = self.get_servicechain_instance(
                context, servicechain_instance_id)
            updated_instance = super(
                NodeCompositionPlugin, self).update_servicechain_instance(
                context, servicechain_instance_id, servicechain_instance)

            if (original_instance['servicechain_specs'] !=
                    updated_instance['servicechain_specs']):
                if len(updated_instance['servicechain_specs']) > 1:
                    raise exc.OneSpecPerInstanceAllowed()
                destroyers = self._get_scheduled_drivers(
                    context, original_instance, 'destroy')
            else:  # Could be classifier update
                updaters = self._get_scheduled_drivers(
                    context, original_instance, 'update')

        if (original_instance['servicechain_specs'] !=
            updated_instance['servicechain_specs']):
            self._destroy_servicechain_nodes(context, destroyers)
            deployers = self._get_scheduled_drivers(
                        context, updated_instance, 'deploy')
            context.deployers = deployers
            context.servicechain_instance = updated_instance
            if not gutils.is_precommit_policy_driver_configured():
                self._deploy_servicechain_nodes(context, deployers)
        else:
            self._update_servicechain_nodes(context, updaters)
        return updated_instance

    @log.log_method_call
    def delete_servicechain_instance(self, context, servicechain_instance_id):
        """Instance deleted.

        When a Servicechain Instance is deleted, all its nodes need to be
        destroyed.
        """
        session = context.session
        with session.begin(subtransactions=True):
            instance = self.get_servicechain_instance(context,
                                                      servicechain_instance_id)
            destroyers = self._get_scheduled_drivers(context, instance,
                                                     'destroy')
        self._destroy_servicechain_nodes(context, destroyers)

        with session.begin(subtransactions=True):
            super(NodeCompositionPlugin, self).delete_servicechain_instance(
                context, servicechain_instance_id)

    @log.log_method_call
    def create_servicechain_node(self, context, servicechain_node):
        session = context.session
        with session.begin(subtransactions=True):
            result = super(NodeCompositionPlugin,
                           self).create_servicechain_node(context,
                                                          servicechain_node)
            self._validate_shared_create(context, result, 'servicechain_node')
        return result

    @log.log_method_call
    def update_servicechain_node(self, context, servicechain_node_id,
                                 servicechain_node):
        """Node Update.

        When a Servicechain Node is updated, all the corresponding instances
        need to be updated as well. This usually results in a node
        reconfiguration.
        """
        session = context.session
        updaters = {}
        with session.begin(subtransactions=True):
            original_sc_node = self.get_servicechain_node(
                context, servicechain_node_id)
            updated_sc_node = super(NodeCompositionPlugin,
                                    self).update_servicechain_node(
                                        context, servicechain_node_id,
                                        servicechain_node)
            self._validate_shared_update(context, original_sc_node,
                                         updated_sc_node, 'servicechain_node')
            instances = self._get_node_instances(context, updated_sc_node)
            for instance in instances:
                node_context = ctx.get_node_driver_context(
                    self, context, instance, updated_sc_node, original_sc_node)
                # TODO(ivar): Validate that the node driver understands the
                # update.
                driver = self.driver_manager.schedule_update(node_context)
                if not driver:
                    raise exc.NoDriverAvailableForAction(
                        action='update', node_id=original_sc_node['id'])
                updaters[instance['id']] = {}
                updaters[instance['id']]['context'] = node_context
                updaters[instance['id']]['driver'] = driver
                updaters[instance['id']]['plumbing_info'] = (
                    driver.get_plumbing_info(node_context))
        # Update the nodes
        for update in updaters.values():
            try:
                update['driver'].update(update['context'])
            except exc.NodeDriverError as ex:
                LOG.error(_LE("Node Update failed, %s"),
                          ex.message)

        return updated_sc_node

    @log.log_method_call
    def get_servicechain_node(self, context, servicechain_node_id,
                              fields = None):
        return self._get_resource(context, 'servicechain_node',
                                  servicechain_node_id, fields)

    @log.log_method_call
    def create_servicechain_spec(self, context, servicechain_spec):
        session = context.session
        with session.begin(subtransactions=True):
            result = super(
                NodeCompositionPlugin, self).create_servicechain_spec(
                    context, servicechain_spec, set_params=False)
            self._validate_shared_create(context, result, 'servicechain_spec')
        return result

    @log.log_method_call
    def update_servicechain_spec(self, context, servicechain_spec_id,
                                 servicechain_spec):
        session = context.session
        with session.begin(subtransactions=True):
            original_sc_spec = self.get_servicechain_spec(
                                         context, servicechain_spec_id)
            updated_sc_spec = super(NodeCompositionPlugin,
                                    self).update_servicechain_spec(
                                        context, servicechain_spec_id,
                                        servicechain_spec, set_params=False)
            self._validate_shared_update(context, original_sc_spec,
                                         updated_sc_spec, 'servicechain_spec')
            # The reference plumber does not support modifying or reordering of
            # nodes in a service chain spec. Disallow update for now
            if (original_sc_spec['nodes'] != updated_sc_spec['nodes'] and
                original_sc_spec['instances']):
                raise exc.InuseSpecNodeUpdateNotAllowed()

        return updated_sc_spec

    @log.log_method_call
    def get_servicechain_spec(self, context,
                              servicechain_spec_id, fields = None):
        return self._get_resource(context, 'servicechain_spec',
                                  servicechain_spec_id, fields)

    @log.log_method_call
    def create_service_profile(self, context, service_profile):
        session = context.session
        with session.begin(subtransactions=True):
            result = super(
                NodeCompositionPlugin, self).create_service_profile(
                    context, service_profile)
            self._validate_shared_create(context, result, 'service_profile')
        return result

    @log.log_method_call
    def update_service_profile(self, context, service_profile_id,
                               service_profile):
        session = context.session
        with session.begin(subtransactions=True):
            original_profile = self.get_service_profile(
                context, service_profile_id)
            updated_profile = super(NodeCompositionPlugin,
                                    self).update_service_profile(
                                        context, service_profile_id,
                                        service_profile)
            self._validate_profile_update(context, original_profile,
                                          updated_profile)
        return updated_profile

    @log.log_method_call
    def get_service_profile(self, context, service_profile_id, fields = None):
        return self._get_resource(context, 'service_profile',
                                  service_profile_id, fields)

    def update_chains_pt_added(self, context, policy_target, instance_id):
        """ Auto scaling function.

        Notify the correct set of node drivers that a new policy target has
        been added to a relevant PTG.
        """
        self._update_chains_pt_modified(context, policy_target, instance_id,
                                        'added')

    def update_chains_pt_removed(self, context, policy_target, instance_id):
        """ Auto scaling function.

        Notify the correct set of node drivers that a new policy target has
        been removed from a relevant PTG.
        """
        self._update_chains_pt_modified(context, policy_target, instance_id,
                                        'removed')

    def update_chains_consumer_added(self, context, policy_target_group,
                                     instance_id):
        """ Auto scaling function.

        Override this method to react to policy target group addition as
        a consumer of a chain.
        """
        self._update_chains_consumer_modified(context, policy_target_group,
                                              instance_id, 'added')

    def policy_target_group_updated(self, context, old_policy_target_group,
                                    current_policy_target_group,
                                    instance_id):
        """ Utility function.
        Override this method to react to policy target group update
        """
        self._policy_target_group_updated(context,
                                          old_policy_target_group,
                                          current_policy_target_group,
                                          instance_id)

    def update_chains_consumer_removed(self, context, policy_target_group,
                                       instance_id):
        """ Auto scaling function.

        Override this method to react to policy target group removed as a
        consumer of a chain
        """
        self._update_chains_consumer_modified(context, policy_target_group,
                                              instance_id, 'removed')

    def _policy_target_group_updated(self, context, old_policy_target_group,
                                     current_policy_target_group,
                                     instance_id):
        updaters = self._get_scheduled_drivers(
                context,
                self.get_servicechain_instance(context, instance_id),
                'update')
        for update in updaters.values():
            try:
                update['driver'].policy_target_group_updated(
                        update['context'],
                        old_policy_target_group,
                        current_policy_target_group)
            except exc.NodeDriverError as ex:
                LOG.error(_LE("Node Update on policy target group modification"
                              " failed, %s"), ex.message)

    def _update_chains_pt_modified(self, context, policy_target, instance_id,
                                   action):
        updaters = self._get_scheduled_drivers(
            context, self.get_servicechain_instance(context, instance_id),
            'update')
        for update in updaters.values():
            try:
                getattr(update['driver'],
                        'update_policy_target_' + action)(
                            update['context'], policy_target)
            except exc.NodeDriverError as ex:
                LOG.error(_LE("Node Update on policy target modification "
                              "failed, %s"), ex.message)

    def _update_chains_consumer_modified(self, context, policy_target_group,
                                         instance_id, action):
        updaters = self._get_scheduled_drivers(
            context, self.get_servicechain_instance(context, instance_id),
            'update')
        for update in updaters.values():
            try:
                getattr(update['driver'],
                        'update_node_consumer_ptg_' + action)(
                            update['context'], policy_target_group)
            except exc.NodeDriverError as ex:
                LOG.error(_LE(
                    "Node Update on policy target group modification "
                    "failed, %s"), ex.message)

    def notify_chain_parameters_updated(self, context,
                                        servicechain_instance_id):
        """Hook for GBP drivers to inform about any updates that affect the SCI

        Notify the correct set of node drivers that some parameter that affects
        the service chain is updated. The updates could be something like
        adding or removing an Allow Rule to the ruleset and may have to be
        enforced in the Firewall Service VM, or it could simply be a
        classifier update.
        """
        sci = self.get_servicechain_instance(context, servicechain_instance_id)
        updaters = self._get_scheduled_drivers(context, sci, 'update')
        for update in updaters.values():
            try:
                getattr(update['driver'],
                        'notify_chain_parameters_updated')(update['context'])
            except exc.NodeDriverError as ex:
                LOG.error(_LE("Node Update on GBP parameter update "
                              "failed, %s"), ex.message)

    def _get_instance_nodes(self, context, instance):
        context = utils.admin_context(context)
        if not instance['servicechain_specs']:
            return []
        specs = self.get_servicechain_spec(
            context, instance['servicechain_specs'][0])
        return self.get_servicechain_nodes(context, {'id': specs['nodes']})

    def _get_node_instances(self, context, node):
        context = utils.admin_context(context)
        specs = self.get_servicechain_specs(
            context, {'id': node['servicechain_specs']})
        result = []
        for spec in specs:
            result.extend(self.get_servicechain_instances(
                context, {'id': spec['instances']}))
        return result

    def _get_scheduled_drivers(self, context, instance, action, nodes=None):
        if nodes is None:
            nodes = self._get_instance_nodes(context, instance)
        result = {}
        func = getattr(self.driver_manager, 'schedule_' + action)
        for node in nodes or []:
            node_context = ctx.get_node_driver_context(
                self, context, instance, node)
            driver = func(node_context)
            if not driver:
                raise exc.NoDriverAvailableForAction(action=action,
                                                     node_id=node['id'])
            result[node['id']] = {}
            result[node['id']]['driver'] = driver
            result[node['id']]['context'] = node_context
            result[node['id']]['plumbing_info'] = driver.get_plumbing_info(
                node_context)
        return result

    def _get_resource(self, context, resource_name, resource_id, fields=None):
        session = context.session
        deployers = {}
        with session.begin(subtransactions=True):
            resource = getattr(super(NodeCompositionPlugin,
                self), 'get_' + resource_name)(context, resource_id)
            if resource_name == 'servicechain_instance':
                if len(resource['servicechain_specs']) > 1:
                    raise exc.OneSpecPerInstanceAllowed()
                try:
                    deployers = self._get_scheduled_drivers(context, resource,
                                                        'get')
                except Exception:
                    LOG.warning(_LW("Failed to get node driver"))

        # Invoke drivers only if status attributes are requested
        if not fields or STATUS_SET.intersection(set(fields)):
            _resource = self._get_resource_status(context, resource_name,
                                                  deployers)
            if _resource:
                updated_status = _resource['status']
                updated_status_details = _resource['status_details']
                if resource['status'] != updated_status or (
                    resource['status_details'] != updated_status_details):
                    new_status = {resource_name:
                              {'status': updated_status,
                               'status_details': updated_status_details}}
                    session = context.session
                    with session.begin(subtransactions=True):
                        getattr(super(NodeCompositionPlugin, self),
                            'update_' + resource_name)(
                             context, resource['id'], new_status)
                    resource['status'] = updated_status
                    resource['status_details'] = updated_status_details
        return self._fields(resource, fields)

    def _get_resource_status(self, context, resource_name, deployers=None):
        """
        Invoke node drivers only for servicechain_instance.
        Node driver should implement get_status api to return status
        and status_details of servicechain_instance
        """
        if resource_name == 'servicechain_instance':
            nodes_status = []
            result = {'status': 'BUILD',
                      'status_details': 'node deployment in progress'}
            if deployers:
                try:
                    for deploy in deployers.values():
                        driver = deploy['driver']
                        nodes_status.append(driver.get_status(
                            deploy['context']))
                    node_status = [node['status'] for node in nodes_status]
                    if 'ERROR' in node_status:
                        result['status'] = 'ERROR'
                        result['status_details'] = 'node deployment failed'
                    elif node_status.count('ACTIVE') == len(
                            deployers.values()):
                        result['status'] = 'ACTIVE'
                        result['status_details'] = 'node deployment completed'
                except Exception as exc:
                    LOG.error(_LE("Failed to get servicechain instance status "
                        "from node driver, Error: %(exc)s"), {'exc': exc})
                    return
                return result
        result = {'status': 'ACTIVE', 'status_details': ''}
        return result

    def _deploy_servicechain_nodes(self, context, deployers):
        self.plumber.plug_services(context, deployers.values())
        for deploy in deployers.values():
            driver = deploy['driver']
            driver.create(deploy['context'])

    def _update_servicechain_nodes(self, context, updaters):
        for update in updaters.values():
            driver = update['driver']
            driver.update(update['context'])

    def _destroy_servicechain_nodes(self, context, destroyers):
        # Actual node disruption
        try:
            for destroy in destroyers.values():
                driver = destroy['driver']
                try:
                    driver.delete(destroy['context'])
                except exc.NodeDriverError:
                    LOG.error(_LE("Node destroy failed, for node %s "),
                              driver['context'].current_node['id'])
                except Exception as e:
                    if db_api.is_retriable(e):
                        with excutils.save_and_reraise_exception():
                            LOG.debug(
                                "Node driver '%(name)s' failed in"
                                " %(method)s, operation will be retried",
                                {'name': driver._name, 'method': 'delete'}
                            )
                    LOG.exception(e)
                finally:
                    self.driver_manager.clear_node_owner(destroy['context'])
        finally:
            self.plumber.unplug_services(context, destroyers.values())

    def _validate_profile_update(self, context, original, updated):
        # Raise if the profile is in use by any instance
        # Ugly one shot query to verify whether the profile is in use
        db = servicechain_db
        query = context.session.query(db.ServiceChainInstance)
        query = query.join(db.InstanceSpecAssociation)
        query = query.join(db.ServiceChainSpec)
        query = query.join(db.SpecNodeAssociation)
        query = query.join(db.ServiceChainNode)
        instance = query.filter(
            db.ServiceChainNode.service_profile_id == original['id']).first()
        if instance:
            raise exc.ServiceProfileInUseByAnInstance(
                profile_id=original['id'], instance_id=instance.id)
        self._validate_shared_update(context, original, updated,
                                     'service_profile')



# Copyright (c) 2014 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
This module holds the data models for the load balancer service plugin.  These
are meant simply as replacement data structures for dictionaries and
SQLAlchemy models.  Using dictionaries as data containers for many components
causes readability issues and does not intuitively give the benefits of what
classes and OO give.  Using SQLAlchemy models as data containers for many
components can become an issue if you do not want to give certain components
access to the database.

These data models do provide methods for instantiation from SQLAlchemy models
and also converting to dictionaries.
"""

from gbpservice.contrib.nfp.configurator.lib import lbv2_constants as l_const


class BaseDataModel(object):

    # NOTE(ihrachys): we could reuse the list to provide a default __init__
    # implementation. That would require handling custom default values though.
    fields = []

    def to_dict(self, **kwargs):
        ret = {}
        for attr in self.__dict__:
            if attr.startswith('_') or not kwargs.get(attr, True):
                continue
            if isinstance(getattr(self, attr), list):
                ret[attr] = []
                for item in self.__dict__[attr]:
                    if isinstance(item, BaseDataModel):
                        ret[attr].append(item.to_dict())
                    else:
                        ret[attr] = item
            elif isinstance(getattr(self, attr), BaseDataModel):
                ret[attr] = self.__dict__[attr].to_dict()
            elif isinstance(self.__dict__[attr], unicode):
                ret[attr.encode('utf8')] = self.__dict__[attr].encode('utf8')
            else:
                ret[attr] = self.__dict__[attr]
        return ret

    def to_api_dict(self, **kwargs):
        return {}

    @classmethod
    def from_dict(cls, model_dict):
        fields = {k: v for k, v in model_dict.items()
                  if k in cls.fields}
        return cls(**fields)

    @property
    def root_loadbalancer(self):
        """Returns the loadbalancer this instance is attached to."""
        if isinstance(self, LoadBalancer):
            lb = self
        elif isinstance(self, Listener):
            lb = self.loadbalancer
        elif isinstance(self, L7Policy):
            lb = self.listener.loadbalancer
        elif isinstance(self, L7Rule):
            lb = self.policy.listener.loadbalancer
        elif isinstance(self, Pool):
            lb = self.loadbalancer
        elif isinstance(self, SNI):
            lb = self.listener.loadbalancer
        else:
            # Pool Member or Health Monitor
            lb = self.pool.loadbalancer
        return lb


# NOTE(brandon-logan) AllocationPool, HostRoute, Subnet, IPAllocation, Port,
# and ProviderResourceAssociation are defined here because there aren't any
# data_models defined in core neutron or neutron services.  Instead of jumping
# through the hoops to create those I've just defined them here.  If ever
# data_models or similar are defined in those packages, those should be used
# instead of these.
class AllocationPool(BaseDataModel):

    fields = ['start', 'end']

    def __init__(self, start=None, end=None):
        self.start = start
        self.end = end


class HostRoute(BaseDataModel):

    fields = ['destination', 'nexthop']

    def __init__(self, destination=None, nexthop=None):
        self.destination = destination
        self.nexthop = nexthop


class Subnet(BaseDataModel):

    fields = ['id', 'name', 'tenant_id', 'network_id', 'ip_version', 'cidr',
              'gateway_ip', 'enable_dhcp', 'ipv6_ra_mode', 'ipv6_address_mode',
              'shared', 'dns_nameservers', 'host_routes', 'allocation_pools',
              'subnetpool_id']

    def __init__(self, id=None, name=None, tenant_id=None, network_id=None,
                 ip_version=None, cidr=None, gateway_ip=None, enable_dhcp=None,
                 ipv6_ra_mode=None, ipv6_address_mode=None, shared=None,
                 dns_nameservers=None, host_routes=None, allocation_pools=None,
                 subnetpool_id=None):
        self.id = id
        self.name = name
        self.tenant_id = tenant_id
        self.network_id = network_id
        self.ip_version = ip_version
        self.cidr = cidr
        self.gateway_ip = gateway_ip
        self.enable_dhcp = enable_dhcp
        self.ipv6_ra_mode = ipv6_ra_mode
        self.ipv6_address_mode = ipv6_address_mode
        self.shared = shared
        self.dns_nameservers = dns_nameservers
        self.host_routes = host_routes
        self.allocation_pools = allocation_pools
        self.subnetpool_id = subnetpool_id

    @classmethod
    def from_dict(cls, model_dict):
        host_routes = model_dict.pop('host_routes', [])
        allocation_pools = model_dict.pop('allocation_pools', [])
        model_dict['host_routes'] = [HostRoute.from_dict(route)
                                     for route in host_routes]
        model_dict['allocation_pools'] = [AllocationPool.from_dict(ap)
                                          for ap in allocation_pools]
        return super(Subnet, cls).from_dict(model_dict)


class IPAllocation(BaseDataModel):

    fields = ['port_id', 'ip_address', 'subnet_id', 'network_id']

    def __init__(self, port_id=None, ip_address=None, subnet_id=None,
                 network_id=None):
        self.port_id = port_id
        self.ip_address = ip_address
        self.subnet_id = subnet_id
        self.network_id = network_id

    @classmethod
    def from_dict(cls, model_dict):
        subnet = model_dict.pop('subnet', None)
        # TODO(blogan): add subnet to __init__.  Can't do it yet because it
        # causes issues with converting SA models into data models.
        instance = super(IPAllocation, cls).from_dict(model_dict)
        setattr(instance, 'subnet', None)
        if subnet:
            setattr(instance, 'subnet', Subnet.from_dict(subnet))
        return instance


class Port(BaseDataModel):

    fields = ['id', 'tenant_id', 'name', 'network_id', 'mac_address',
              'admin_state_up', 'status', 'device_id', 'device_owner',
              'fixed_ips']

    def __init__(self, id=None, tenant_id=None, name=None, network_id=None,
                 mac_address=None, admin_state_up=None, status=None,
                 device_id=None, device_owner=None, fixed_ips=None):
        self.id = id
        self.tenant_id = tenant_id
        self.name = name
        self.network_id = network_id
        self.mac_address = mac_address
        self.admin_state_up = admin_state_up
        self.status = status
        self.device_id = device_id
        self.device_owner = device_owner
        self.fixed_ips = fixed_ips or []

    @classmethod
    def from_dict(cls, model_dict):
        fixed_ips = model_dict.pop('fixed_ips', [])
        model_dict['fixed_ips'] = [IPAllocation.from_dict(fixed_ip)
                                   for fixed_ip in fixed_ips]
        return super(Port, cls).from_dict(model_dict)


class ProviderResourceAssociation(BaseDataModel):

    fields = ['provider_name', 'resource_id']

    def __init__(self, provider_name=None, resource_id=None):
        self.provider_name = provider_name
        self.resource_id = resource_id

    @classmethod
    def from_dict(cls, model_dict):
        device_driver = model_dict.pop('device_driver', None)
        instance = super(ProviderResourceAssociation, cls).from_dict(
            model_dict)
        setattr(instance, 'device_driver', device_driver)
        return instance


class SessionPersistence(BaseDataModel):

    fields = ['pool_id', 'type', 'cookie_name', 'pool']

    def __init__(self, pool_id=None, type=None, cookie_name=None,
                 pool=None):
        self.pool_id = pool_id
        self.type = type
        self.cookie_name = cookie_name
        self.pool = pool

    def to_api_dict(self):
        return super(SessionPersistence, self).to_dict(pool=False,
                                                       pool_id=False)

    @classmethod
    def from_dict(cls, model_dict):
        pool = model_dict.pop('pool', None)
        if pool:
            model_dict['pool'] = Pool.from_dict(
                pool)
        return super(SessionPersistence, cls).from_dict(model_dict)


class LoadBalancerStatistics(BaseDataModel):

    fields = ['loadbalancer_id', 'bytes_in', 'bytes_out', 'active_connections',
              'total_connections', 'loadbalancer']

    def __init__(self, loadbalancer_id=None, bytes_in=None, bytes_out=None,
                 active_connections=None, total_connections=None,
                 loadbalancer=None):
        self.loadbalancer_id = loadbalancer_id
        self.bytes_in = bytes_in
        self.bytes_out = bytes_out
        self.active_connections = active_connections
        self.total_connections = total_connections
        self.loadbalancer = loadbalancer

    def to_api_dict(self):
        return super(LoadBalancerStatistics, self).to_dict(
            loadbalancer_id=False, loadbalancer=False)


class HealthMonitor(BaseDataModel):

    fields = ['id', 'tenant_id', 'type', 'delay', 'timeout', 'max_retries',
              'http_method', 'url_path', 'expected_codes',
              'provisioning_status', 'admin_state_up', 'pool', 'name']

    def __init__(self, id=None, tenant_id=None, type=None, delay=None,
                 timeout=None, max_retries=None, http_method=None,
                 url_path=None, expected_codes=None, provisioning_status=None,
                 admin_state_up=None, pool=None, name=None):
        self.id = id
        self.tenant_id = tenant_id
        self.type = type
        self.delay = delay
        self.timeout = timeout
        self.max_retries = max_retries
        self.http_method = http_method
        self.url_path = url_path
        self.expected_codes = expected_codes
        self.provisioning_status = provisioning_status
        self.admin_state_up = admin_state_up
        self.pool = pool
        self.name = name

    def attached_to_loadbalancer(self):
        return bool(self.pool and self.pool.loadbalancer)

    def to_api_dict(self):
        ret_dict = super(HealthMonitor, self).to_dict(
            provisioning_status=False, pool=False)
        ret_dict['pools'] = []
        if self.pool:
            ret_dict['pools'].append({'id': self.pool.id})
        if self.type in [l_const.HEALTH_MONITOR_TCP,
                         l_const.HEALTH_MONITOR_PING]:
            ret_dict.pop('http_method')
            ret_dict.pop('url_path')
            ret_dict.pop('expected_codes')
        return ret_dict

    @classmethod
    def from_dict(cls, model_dict):
        pool = model_dict.pop('pool', None)
        if pool:
            model_dict['pool'] = Pool.from_dict(
                pool)
        return super(HealthMonitor, cls).from_dict(model_dict)


class Pool(BaseDataModel):

    fields = ['id', 'tenant_id', 'name', 'description', 'healthmonitor_id',
              'protocol', 'lb_algorithm', 'admin_state_up', 'operating_status',
              'provisioning_status', 'members', 'healthmonitor',
              'session_persistence', 'loadbalancer_id', 'loadbalancer',
              'listener', 'listeners', 'l7_policies']

    # Map deprecated attribute names to new ones.
    attr_mapping = {'sessionpersistence': 'session_persistence'}

    def __init__(self, id=None, tenant_id=None, name=None, description=None,
                 healthmonitor_id=None, protocol=None, lb_algorithm=None,
                 admin_state_up=None, operating_status=None,
                 provisioning_status=None, members=None, healthmonitor=None,
                 session_persistence=None, loadbalancer_id=None,
                 loadbalancer=None, listener=None, listeners=None,
                 l7_policies=None):
        self.id = id
        self.tenant_id = tenant_id
        self.name = name
        self.description = description
        self.healthmonitor_id = healthmonitor_id
        self.protocol = protocol
        self.lb_algorithm = lb_algorithm
        self.admin_state_up = admin_state_up
        self.operating_status = operating_status
        self.provisioning_status = provisioning_status
        self.members = members or []
        self.healthmonitor = healthmonitor
        self.session_persistence = session_persistence
        # NOTE(eezhova): Old attribute name is kept for backwards
        # compatibility with out-of-tree drivers.
        self.sessionpersistence = self.session_persistence
        self.loadbalancer_id = loadbalancer_id
        self.loadbalancer = loadbalancer
        self.listener = listener
        self.listeners = listeners or []
        self.l7_policies = l7_policies or []

    def attached_to_loadbalancer(self):
        return bool(self.loadbalancer)

    def to_api_dict(self):
        ret_dict = super(Pool, self).to_dict(
            provisioning_status=False, operating_status=False,
            healthmonitor=False, session_persistence=False,
            loadbalancer_id=False, loadbalancer=False, listener_id=False)
        ret_dict['loadbalancers'] = []
        if self.loadbalancer:
            ret_dict['loadbalancers'].append({'id': self.loadbalancer.id})
        ret_dict['session_persistence'] = None
        if self.session_persistence:
            ret_dict['session_persistence'] = (
                self.session_persistence.to_api_dict())
        ret_dict['members'] = [{'id': member.id} for member in self.members]
        ret_dict['listeners'] = [{'id': listener.id}
                                 for listener in self.listeners]
        if self.listener:
            ret_dict['listener_id'] = self.listener.id
        else:
            ret_dict['listener_id'] = None
        ret_dict['l7_policies'] = [{'id': l7_policy.id}
                                   for l7_policy in self.l7_policies]
        return ret_dict

    @classmethod
    def from_dict(cls, model_dict):
        healthmonitor = model_dict.pop('healthmonitor', None)
        session_persistence = model_dict.pop('session_persistence', None)
        model_dict.pop('sessionpersistence', None)
        loadbalancer = model_dict.pop('loadbalancer', None)
        members = model_dict.pop('members', [])
        model_dict['members'] = [Member.from_dict(member)
                                 for member in members]
        listeners = model_dict.pop('listeners', [])
        model_dict['listeners'] = [Listener.from_dict(listener)
                                   for listener in listeners]
        l7_policies = model_dict.pop('l7_policies', [])
        model_dict['l7_policies'] = [L7Policy.from_dict(policy)
                                     for policy in l7_policies]

        # handle old attribute for out of tree drivers
        listener = model_dict.pop('listener', None)
        if listener:
            model_dict['listener'] = Listener.from_dict(listener)

        if healthmonitor:
            model_dict['healthmonitor'] = HealthMonitor.from_dict(
                healthmonitor)
        if session_persistence:
            model_dict['session_persistence'] = SessionPersistence.from_dict(
                session_persistence)
        if loadbalancer:
            model_dict['loadbalancer'] = LoadBalancer.from_dict(loadbalancer)
        return super(Pool, cls).from_dict(model_dict)


class Member(BaseDataModel):

    fields = ['id', 'tenant_id', 'pool_id', 'address', 'protocol_port',
              'weight', 'admin_state_up', 'subnet_id', 'operating_status',
              'provisioning_status', 'pool', 'name']

    def __init__(self, id=None, tenant_id=None, pool_id=None, address=None,
                 protocol_port=None, weight=None, admin_state_up=None,
                 subnet_id=None, operating_status=None,
                 provisioning_status=None, pool=None, name=None):
        self.id = id
        self.tenant_id = tenant_id
        self.pool_id = pool_id
        self.address = address
        self.protocol_port = protocol_port
        self.weight = weight
        self.admin_state_up = admin_state_up
        self.subnet_id = subnet_id
        self.operating_status = operating_status
        self.provisioning_status = provisioning_status
        self.pool = pool
        self.name = name

    def attached_to_loadbalancer(self):
        return bool(self.pool and self.pool.loadbalancer)

    def to_api_dict(self):
        return super(Member, self).to_dict(
            provisioning_status=False, operating_status=False, pool=False)

    @classmethod
    def from_dict(cls, model_dict):
        pool = model_dict.pop('pool', None)
        if pool:
            model_dict['pool'] = Pool.from_dict(
                pool)
        return super(Member, cls).from_dict(model_dict)


class SNI(BaseDataModel):

    fields = ['listener_id', 'tls_container_id', 'position', 'listener']

    def __init__(self, listener_id=None, tls_container_id=None,
                 position=None, listener=None):
        self.listener_id = listener_id
        self.tls_container_id = tls_container_id
        self.position = position
        self.listener = listener

    def attached_to_loadbalancer(self):
        return bool(self.listener and self.listener.loadbalancer)

    def to_api_dict(self):
        return super(SNI, self).to_dict(listener=False)


class TLSContainer(BaseDataModel):

    fields = ['id', 'certificate', 'private_key', 'passphrase',
              'intermediates', 'primary_cn']

    def __init__(self, id=None, certificate=None, private_key=None,
                 passphrase=None, intermediates=None, primary_cn=None):
        self.id = id
        self.certificate = certificate
        self.private_key = private_key
        self.passphrase = passphrase
        self.intermediates = intermediates
        self.primary_cn = primary_cn


class L7Rule(BaseDataModel):

    fields = ['id', 'tenant_id', 'l7policy_id', 'type', 'compare_type',
              'invert', 'key', 'value', 'provisioning_status',
              'admin_state_up', 'policy']

    def __init__(self, id=None, tenant_id=None,
                 l7policy_id=None, type=None, compare_type=None, invert=None,
                 key=None, value=None, provisioning_status=None,
                 admin_state_up=None, policy=None):
        self.id = id
        self.tenant_id = tenant_id
        self.l7policy_id = l7policy_id
        self.type = type
        self.compare_type = compare_type
        self.invert = invert
        self.key = key
        self.value = value
        self.provisioning_status = provisioning_status
        self.admin_state_up = admin_state_up
        self.policy = policy

    def attached_to_loadbalancer(self):
        return bool(self.policy.listener.loadbalancer)

    def to_api_dict(self):
        ret_dict = super(L7Rule, self).to_dict(
            provisioning_status=False,
            policy=False, l7policy_id=False)
        ret_dict['policies'] = []
        if self.policy:
            ret_dict['policies'].append({'id': self.policy.id})
        return ret_dict

    @classmethod
    def from_dict(cls, model_dict):
        policy = model_dict.pop('policy', None)
        if policy:
            model_dict['policy'] = L7Policy.from_dict(policy)
        return super(L7Rule, cls).from_dict(model_dict)


class L7Policy(BaseDataModel):

    fields = ['id', 'tenant_id', 'name', 'description', 'listener_id',
              'action', 'redirect_pool_id', 'redirect_url', 'position',
              'admin_state_up', 'provisioning_status', 'listener', 'rules',
              'redirect_pool']

    def __init__(self, id=None, tenant_id=None, name=None, description=None,
                 listener_id=None, action=None, redirect_pool_id=None,
                 redirect_url=None, position=None,
                 admin_state_up=None, provisioning_status=None,
                 listener=None, rules=None, redirect_pool=None):
        self.id = id
        self.tenant_id = tenant_id
        self.name = name
        self.description = description
        self.listener_id = listener_id
        self.action = action
        self.redirect_pool_id = redirect_pool_id
        self.redirect_pool = redirect_pool
        self.redirect_url = redirect_url
        self.position = position
        self.admin_state_up = admin_state_up
        self.provisioning_status = provisioning_status
        self.listener = listener
        self.rules = rules or []

    def attached_to_loadbalancer(self):
        return bool(self.listener.loadbalancer)

    def to_api_dict(self):
        ret_dict = super(L7Policy, self).to_dict(
            listener=False, listener_id=False,
            provisioning_status=False, redirect_pool=False)
        ret_dict['listeners'] = []
        if self.listener:
            ret_dict['listeners'].append({'id': self.listener.id})
        ret_dict['rules'] = [{'id': rule.id} for rule in self.rules]
        return ret_dict

    @classmethod
    def from_dict(cls, model_dict):
        listener = model_dict.pop('listener', None)
        redirect_pool = model_dict.pop('redirect_pool', None)
        rules = model_dict.pop('rules', [])
        if listener:
            model_dict['listener'] = Listener.from_dict(listener)
        if redirect_pool:
            model_dict['redirect_pool'] = Pool.from_dict(redirect_pool)
        model_dict['rules'] = [L7Rule.from_dict(rule)
                               for rule in rules]
        return super(L7Policy, cls).from_dict(model_dict)


class Listener(BaseDataModel):

    fields = ['id', 'tenant_id', 'name', 'description', 'default_pool_id',
              'loadbalancer_id', 'protocol', 'default_tls_container_id',
              'sni_containers', 'protocol_port', 'connection_limit',
              'admin_state_up', 'provisioning_status', 'operating_status',
              'default_pool', 'loadbalancer', 'l7_policies']

    def __init__(self, id=None, tenant_id=None, name=None, description=None,
                 default_pool_id=None, loadbalancer_id=None, protocol=None,
                 default_tls_container_id=None, sni_containers=None,
                 protocol_port=None, connection_limit=None,
                 admin_state_up=None, provisioning_status=None,
                 operating_status=None, default_pool=None, loadbalancer=None,
                 l7_policies=None):
        self.id = id
        self.tenant_id = tenant_id
        self.name = name
        self.description = description
        self.default_pool_id = default_pool_id
        self.loadbalancer_id = loadbalancer_id
        self.protocol = protocol
        self.default_tls_container_id = default_tls_container_id
        self.sni_containers = sni_containers or []
        self.protocol_port = protocol_port
        self.connection_limit = connection_limit
        self.admin_state_up = admin_state_up
        self.operating_status = operating_status
        self.provisioning_status = provisioning_status
        self.default_pool = default_pool
        self.loadbalancer = loadbalancer
        self.l7_policies = l7_policies or []

    def attached_to_loadbalancer(self):
        return bool(self.loadbalancer)

    def to_api_dict(self):
        ret_dict = super(Listener, self).to_dict(
            loadbalancer=False, loadbalancer_id=False, default_pool=False,
            operating_status=False, provisioning_status=False,
            sni_containers=False)
        # NOTE(blogan): Returning a list to future proof for M:N objects
        # that are not yet implemented.
        ret_dict['loadbalancers'] = []
        if self.loadbalancer:
            ret_dict['loadbalancers'].append({'id': self.loadbalancer.id})
        ret_dict['sni_container_refs'] = [container.tls_container_id
                                          for container in self.sni_containers]
        ret_dict['default_tls_container_ref'] = self.default_tls_container_id
        ret_dict['l7_policies'] = [{'id': l7_policy.id}
                                   for l7_policy in self.l7_policies]
        return ret_dict

    @classmethod
    def from_dict(cls, model_dict):
        default_pool = model_dict.pop('default_pool', None)
        loadbalancer = model_dict.pop('loadbalancer', None)
        sni_containers = model_dict.pop('sni_containers', [])
        model_dict['sni_containers'] = [SNI.from_dict(sni)
                                        for sni in sni_containers]
        l7_policies = model_dict.pop('l7_policies', [])
        if default_pool:
            model_dict['default_pool'] = Pool.from_dict(default_pool)
        if loadbalancer:
            model_dict['loadbalancer'] = LoadBalancer.from_dict(loadbalancer)
        model_dict['l7_policies'] = [L7Policy.from_dict(policy)
                                     for policy in l7_policies]
        return super(Listener, cls).from_dict(model_dict)


class LoadBalancer(BaseDataModel):

    fields = ['id', 'tenant_id', 'name', 'description', 'vip_subnet_id',
              'vip_port_id', 'vip_address', 'provisioning_status',
              'operating_status', 'admin_state_up', 'vip_port', 'stats',
              'provider', 'listeners', 'pools', 'flavor_id']

    def __init__(self, id=None, tenant_id=None, name=None, description=None,
                 vip_subnet_id=None, vip_port_id=None, vip_address=None,
                 provisioning_status=None, operating_status=None,
                 admin_state_up=None, vip_port=None, stats=None,
                 provider=None, listeners=None, pools=None, flavor_id=None):
        self.id = id
        self.tenant_id = tenant_id
        self.name = name
        self.description = description
        self.vip_subnet_id = vip_subnet_id
        self.vip_port_id = vip_port_id
        self.vip_address = vip_address
        self.operating_status = operating_status
        self.provisioning_status = provisioning_status
        self.admin_state_up = admin_state_up
        self.vip_port = vip_port
        self.stats = stats
        self.provider = provider
        self.listeners = listeners or []
        self.flavor_id = flavor_id
        self.pools = pools or []

    def attached_to_loadbalancer(self):
        return True

    def to_api_dict(self):
        ret_dict = super(LoadBalancer, self).to_dict(
            vip_port=False, stats=False, listeners=False)
        ret_dict['listeners'] = [{'id': listener.id}
                                 for listener in self.listeners]
        ret_dict['pools'] = [{'id': pool.id} for pool in self.pools]
        if self.provider:
            ret_dict['provider'] = self.provider.provider_name

        if not self.flavor_id:
            del ret_dict['flavor_id']

        return ret_dict

    @classmethod
    def from_dict(cls, model_dict):
        listeners = model_dict.pop('listeners', [])
        pools = model_dict.pop('pools', [])
        vip_port = model_dict.pop('vip_port', None)
        provider = model_dict.pop('provider', None)
        model_dict.pop('stats', None)
        model_dict['listeners'] = [Listener.from_dict(listener)
                                   for listener in listeners]
        model_dict['pools'] = [Pool.from_dict(pool)
                               for pool in pools]
        if vip_port:
            model_dict['vip_port'] = Port.from_dict(vip_port)
        if provider:
            model_dict['provider'] = ProviderResourceAssociation.from_dict(
                provider)
        return super(LoadBalancer, cls).from_dict(model_dict)


NAME_TO_DATA_MODEL_MAP = {
    "loadbalancer": LoadBalancer,
    "healthmonitor": HealthMonitor,
    "listener": Listener,
    "sni": SNI,
    "pool": Pool,
    "member": Member,
    "loadbalancerstatistics": LoadBalancerStatistics,
    "sessionpersistence": SessionPersistence,
    "ipallocation": IPAllocation,
    "port": Port,
    "providerresourceassociation": ProviderResourceAssociation
}



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg as oslo_config

nfp_configurator_extra_opts = [
    oslo_config.StrOpt('log_forward_ip_address',
                       default='', help='Log collector host IP address'),
    oslo_config.IntOpt('log_forward_port',
                       default='514', help='Log collector port number'),
    oslo_config.StrOpt('log_level',
                       default='debug',
                       help='Log level info/error/debug/warning')]

oslo_config.CONF.register_opts(nfp_configurator_extra_opts, "configurator")


nfp_configurator_config_drivers_opts = [
    oslo_config.ListOpt(
        'drivers',
        default=['gbpservice.contrib.nfp.configurator.drivers'],
        help='List of config driver directories')]

oslo_config.CONF.register_opts(nfp_configurator_config_drivers_opts,
                               "CONFIG_DRIVERS")

nfp_configurator_config_agents_opts = [
    oslo_config.ListOpt(
        'agents',
        default=['gbpservice.contrib.nfp.configurator.agents'],
        help='Config agents directory')]

oslo_config.CONF.register_opts(nfp_configurator_config_agents_opts,
                               "CONFIG_AGENTS")



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from gbpclient.v2_0 import client as gbp_client
from keystoneauth1.identity import v2
from keystoneauth1.identity import v3
from keystoneauth1 import session
from keystoneclient.v2_0 import client as identity_client
from keystoneclient.v3 import client as keyclientv3
from neutronclient.v2_0 import client as neutron_client
from novaclient import client as nova_client
from novaclient import exceptions as nova_exc

from gbpservice._i18n import _
from gbpservice.nfp.core import log as nfp_logging
LOG = nfp_logging.getLogger(__name__)


class OpenstackApi(object):
    """Initializes common attributes for openstack client drivers."""

    def __init__(self, config, username=None,
                 password=None, tenant_name=None):
        self.nova_version = '2'
        self.config = config
        self.identity_service = ("%s://%s:%d/%s/" %
                                 (config.nfp_keystone_authtoken.auth_protocol,
                                  config.nfp_keystone_authtoken.auth_host,
                                  config.nfp_keystone_authtoken.auth_port,
                                  config.nfp_keystone_authtoken.auth_version))
        self.network_service = ("%s://%s:%d/" %
                                (config.nfp_keystone_authtoken.auth_protocol,
                                 config.nfp_keystone_authtoken.auth_host,
                                 config.bind_port))
        self.username = username or config.nfp_keystone_authtoken.admin_user
        self.password = password or (
            config.nfp_keystone_authtoken.admin_password)
        self.tenant_name = (tenant_name or
                            config.nfp_keystone_authtoken.admin_tenant_name)
        self.token = None
        self.admin_tenant_id = None


class KeystoneClient(OpenstackApi):
    """ Keystone Client Apis for orchestrator. """

    def get_admin_token(self):
        try:
            admin_token = self.get_scoped_keystone_token(
                self.config.nfp_keystone_authtoken.admin_user,
                self.config.nfp_keystone_authtoken.admin_password,
                self.config.nfp_keystone_authtoken.admin_tenant_name)
        except Exception as ex:
            err = ("Failed to obtain user token. Error: %s" % ex)
            LOG.error(err)
            raise Exception(err)

        return admin_token

    def get_scoped_keystone_token(self, user, password, tenant_name,
                                  tenant_id=None):
        """ Get a scoped token from Openstack Keystone service.

        A scoped token is bound to the specific tenant.

        :param user: User name
        :param password: Password
        :param tenantName: Tenant name

        :return: A scoped token or None if unable to get
        """
        if not (tenant_name or tenant_id):
            err = "Tenant Not specified for getting a scoped token"
            LOG.error(err)
            raise Exception(err)
        try:
            auth = v2.Password(username=user,
                           password=password,
                           tenant_name=tenant_name,
                           auth_url=self.identity_service)
            sess = session.Session(auth=auth)
            scoped_token = sess.get_token(auth=auth)
        except Exception as err:
            err = ("Failed to get token from"
                   " Openstack Keystone service"
                   " KeyError :: %s" % (err))
            self.config.nfp_keystone_authtoken.auth_port,
            LOG.error(err)
            raise Exception(err)
        else:
            return scoped_token

    def get_admin_tenant_id(self, token):
        if not self.admin_tenant_id:
            self.admin_tenant_id = self.get_tenant_id(
                token,
                self.config.nfp_keystone_authtoken.admin_tenant_name)

        return self.admin_tenant_id

    def get_tenant_id(self, token, tenant_name):
        """ Get the tenant UUID associated to tenant name

        :param token: A scoped token
        :param tenant: Tenant name

        :return: Tenant UUID
        """
        try:
            keystone = self._get_v2_keystone_admin_client()
            tenant = keystone.tenants.find(name=tenant_name)
            return tenant.id
        except Exception as ex:
            err = ("Failed to read tenant UUID from"
                   " tenant_name %s."
                   " Error :: %s" % (tenant_name, ex))
            LOG.error(err)
            raise Exception(err)
        err = 'No tenant with name "%s" found in keystone db' % tenant_name
        LOG.error(err)
        raise Exception(err)

    def _get_v2_keystone_admin_client(self):
        """ Returns keystone v2 client with admin credentials
            Using this client one can perform CRUD operations over
            keystone resources.
        """
        keystone_conf = self.config.nfp_keystone_authtoken
        auth = v2.Password(username=keystone_conf.admin_user,
                           password=keystone_conf.admin_password,
                           tenant_name=keystone_conf.admin_tenant_name,
                           auth_url=self.identity_service)
        sess = session.Session(auth=auth)
        v2client = identity_client.Client(session=sess)
        return v2client

    def _get_v3_keystone_admin_client(self):
        """ Returns keystone v3 client with admin credentials
            Using this client one can perform CRUD operations over
            keystone resources.
        """
        keystone_conf = self.config.nfp_keystone_authtoken
        v3_auth_url = ('%s://%s:%s/%s/' % (
            keystone_conf.auth_protocol, keystone_conf.auth_host,
            keystone_conf.auth_port, self.config.heat_driver.keystone_version))
        auth = v3.Password(auth_url=v3_auth_url,
                           user_domain_name='Default',
                           username=keystone_conf.admin_user,
                           password=keystone_conf.admin_password,
                           project_domain_name="Default",
                           project_name=keystone_conf.admin_tenant_name)
        sess = session.Session(auth=auth)
        v3client = keyclientv3.Client(session=sess)
        return v3client


class NovaClient(OpenstackApi):
    """ Nova Client Api driver. """

    def get_image_id(self, token, tenant_id, image_name):
        """ Get the image UUID associated to image name

        :param token: A scoped token
        :param tenant_id: Tenant UUID
        :param image_name: Image name

        :return: Image UUID
        """
        try:
            nova = nova_client.Client(self.nova_version, auth_token=token,
                                      tenant_id=tenant_id,
                                      auth_url=self.identity_service)
            image = nova.images.find(name=image_name)
            return image.id
        except Exception as ex:
            err = ("Failed to get image id from image name %s: %s" % (
                image_name, ex))
            LOG.error(err)
            raise Exception(err)

    def get_image_metadata(self, token, tenant_id, image_name):
        """ Get the image UUID associated to image name

        :param token: A scoped token
        :param tenant_id: Tenant UUID
        :param image_name: Image name

        :return: Image UUID
        """
        try:
            nova = nova_client.Client(self.nova_version, auth_token=token,
                                      tenant_id=tenant_id,
                                      auth_url=self.identity_service)
            image = nova.images.find(name=image_name)
            return image.metadata
        except Exception as ex:
            err = ("Failed to get image metadata from image name %s: %s" % (
                image_name, ex))
            LOG.error(err)
            raise Exception(err)

    def get_flavor_id(self, token, tenant_id, flavor_name):
        """ Get the flavor UUID associated to flavor name

        :param token: A scoped token
        :param tenant_id: Tenant UUID
        :param flavor_name: Flavor name

        :return: Flavor UUID or None if not found
        """
        try:
            nova = nova_client.Client(self.nova_version, auth_token=token,
                                      tenant_id=tenant_id,
                                      auth_url=self.identity_service)
            flavor = nova.flavors.find(name=flavor_name)
            return flavor.id
        except Exception as ex:
            err = ("Failed to get flavor id from flavor name %s: %s" % (
                flavor_name, ex))
            LOG.error(err)
            raise Exception(err)

    def get_instance(self, token, tenant_id, instance_id):
        """ Get instance details

        :param token: A scoped_token
        :param tenant_id: Tenant UUID
        :param instance_id: Instance UUID

        :return: Instance details

        """
        try:
            nova = nova_client.Client(self.nova_version, auth_token=token,
                                      tenant_id=tenant_id,
                                      auth_url=self.identity_service)
            instance = nova.servers.get(instance_id)
            if instance:
                return instance.to_dict()
            raise Exception(_("No instance with id %(id)s "
                              "found in db for tenant %(tenant)s")
                            % {'id': instance_id, 'tenant': tenant_id})
        except Exception as ex:
            err = ("Failed to read instance information from"
                   " Openstack Nova service's response"
                   " KeyError :: %s" % (ex))
            LOG.error(err)
            raise Exception(err)

    def get_keypair(self, token, tenant_id, keypair_name):
        """ Get Nova keypair details

        :param token: A scoped_token
        :param tenant_id: Tenant UUID
        :param keypair_name: Nova keypair name

        :return: Nova keypair details

        """
        tenant_id = str(tenant_id)
        try:
            nova = nova_client.Client(self.nova_version, auth_token=token,
                                      tenant_id=tenant_id,
                                      auth_url=self.identity_service)
            keypair = nova.keypairs.find(name=keypair_name)
            return keypair.to_dict()
        except Exception as ex:
            err = ("Failed to read keypair information from"
                   " Openstack Nova service's response."
                   " %s" % ex)
            LOG.error(err)
            raise Exception(err)

    def attach_interface(self, token, tenant_id, instance_id, port_id):
        """ Attaches a port to already created instance
        :param token: A scoped token
        :param tenant_id: Tenant UUID
        :param instance_id: UUID of the instance
        :param port_id: Port UUID
        """
        try:
            nova = nova_client.Client(self.nova_version, auth_token=token,
                                      tenant_id=tenant_id,
                                      auth_url=self.identity_service)
            instance = nova.servers.interface_attach(instance_id, port_id,
                                                     None, None)
            return instance
        except Exception as ex:
            err = ("Failed to attach interface %s to instance"
                   " %s  %s" % (port_id, instance_id, ex))
            LOG.error(err)
            raise Exception(err)

    def detach_interface(self, token, tenant_id, instance_id, port_id):
        """ Detaches a port to already created instance
        :param token: A scoped token
        :param tenant_id: Tenant UUID
        :param instance_id: UUID of the instance
        :param port_id: Port UUID
        """
        try:
            nova = nova_client.Client(self.nova_version, auth_token=token,
                                      tenant_id=tenant_id,
                                      auth_url=self.identity_service)
            instance = nova.servers.interface_detach(instance_id, port_id)
            return instance
        except Exception as ex:
            err = ("Failed to detach interface %s from instance"
                   " %s  %s" % (port_id, instance_id, ex))
            LOG.error(err)
            raise Exception(err)

    def delete_instance(self, token, tenant_id, instance_id):
        """ Delete the instance

        :param token: A scoped token
        :param tenant_id: Tenant UUID
        :param instance_id: Instance UUID

        """
        try:
            nova = nova_client.Client(self.nova_version, auth_token=token,
                                      tenant_id=tenant_id,
                                      auth_url=self.identity_service)
            nova.servers.delete(instance_id)
        except Exception as ex:
            err = ("Failed to delete instance"
                   " %s  %s" % (instance_id, ex))
            LOG.error(err)
            raise Exception(err)

    def get_instances(self, token, filters=None):
        """ List instances

        :param token: A scoped_token
        :param filters: Parameters for list filter
        example for filter: {}, tenant_id is mandatory

        :return: instance List

        """
        if (
            not filters or
            type(filters) != dict or
            'tenant_id' not in filters
        ):
            err = ("Failed to process get_instances,"
                   " filters(type: dict) with tenant_id is mandatory")
            LOG.error(err)
            raise Exception(err)

        tenant_id = filters.get('tenant_id')
        try:
            nova = nova_client.Client(self.nova_version, auth_token=token,
                                      tenant_id=tenant_id,
                                      auth_url=self.identity_service)
            instances = nova.servers.list(search_opts=filters)
            data = [instance.to_dict() for instance in instances]
            return data
        except Exception as ex:
            err = ("Failed to list instances under tenant"
                   " %s  %s" % (tenant_id, ex))
            LOG.error(err)
            raise Exception(err)

    def delete_affinity_group(self, token, tenant_id, nf_id):
        """ Deletes a server group

        :param token: A scoped token
        :param tenant_id: Tenant UUID
        :param nf_id: Network Function UUID

        Returns: None

        """

        nova_version = 2.15
        nova = nova_client.Client(nova_version, auth_token=token,
                                  tenant_id=tenant_id,
                                  auth_url=self.identity_service)

        try:
            affinity_group = nova.server_groups.find(name=nf_id)
            affinity_group_id = affinity_group.to_dict()['id']
            nova.server_groups.delete(affinity_group_id)
            msg = ("Successfully deleted Nova Server Anti-Affinity "
                   "Group: %s" % nf_id)
            LOG.info(msg)
        except nova_exc.NotFound:
            pass
        except Exception as err:
            msg = ("Failed to delete Nova Server Anti-Affinity Group "
                   "with name %s. Error: %s" % (nf_id, err))
            LOG.error(msg)

    def create_affinity_group(self, token, tenant_id, nf_id):
        """ Creates a server group

        :param token: A scoped token
        :param tenant_id: Tenant UUID
        :param nf_id: Network Function UUID

        Returns: Nova server-group json object

        """

        nova_version = 2.15
        kwargs = dict(name=nf_id, policies=['soft-anti-affinity'])
        nova = nova_client.Client(nova_version, auth_token=token,
                                  tenant_id=tenant_id,
                                  auth_url=self.identity_service)

        try:
            affinity_group = nova.server_groups.create(**kwargs)
            affinity_group_id = affinity_group.to_dict()['id']
            msg = ("Successfully created Nova Server Anti-Affinity "
                   "Group: %s" % nf_id)
            LOG.info(msg)
            return affinity_group_id
        except Exception as err:
            msg = ("Failed to create Nova Server Anti-Affinity Group. "
                   "Error: %s" % err)
            LOG.error(msg)
            return None

    def create_instance(self, token, tenant_id, image_id, flavor,
                        nw_port_id_list, name, volume_support,
                        volume_size, secgroup_name=None,
                        metadata=None, files=None, config_drive=False,
                        userdata=None, key_name='', server_grp_id=None,
                        ):
        """ Launch a VM with given details

        :param token: A scoped token
        :param tenant_id: Tenant UUID
        :param image_id: Image UUID
        :param flavor: Flavor name
        :param nw_port_id_list: Network UUID and port UUID list
        :param name: Service istance name
        :param secgroup_name: Nova security group name
        :param metadata: metadata key-value pairs
        :param files: List of files to be copied.
        :example files: [{"dst": <detination_path_string>,
                          "src": <file_contents>}]
        :param userdata: user data to pass to be exposed by the metadata
                      server this can be a file type object as well or a
                      string
        :param key_name: Nova keypair name
        :param server_grp_id: Nova server group UUID
        :param volume_support: volume support to launch instance
        :param volume_size: cinder volume size in GB
        :return: VM instance UUID

        """
        try:
            if files:
                file_dict = {}
                for _file in files:
                    with open(_file["src"]) as config_file:
                        data = config_file.read()
                        config_drive = True
                        file_dict.update({_file["dst"]: data})
                files = file_dict
        except Exception as e:
            msg = (
                "Failed while reading file: %r " % e)
            LOG.error(msg)
            raise e

        kwargs = dict()
        if volume_support:
            block_device_mapping_v2 = [
                {
                    "boot_index": "1",
                    "uuid": image_id,
                    "source_type": "image",
                    "volume_size": volume_size,
                    "destination_type": "volume",
                    "delete_on_termination": True
                }
            ]
            kwargs.update(block_device_mapping_v2=block_device_mapping_v2)

        if server_grp_id:
            kwargs.update(scheduler_hints={"group": server_grp_id})
        if key_name != '':
            kwargs.update(key_name=key_name)
        if config_drive is True:
            kwargs.update(config_drive=True)
        if userdata is not None and \
                (type(userdata) is str or type(userdata) is file):
            kwargs.update(userdata=userdata)
        if metadata is not None and type(metadata) is dict and metadata != {}:
            kwargs.update(meta=metadata)
        if files is not None:
            kwargs.update(files=files)
        if nw_port_id_list:
            nics = [{"port-id": entry.get("port"), "net-id": entry.get("uuid"),
                     "v4-fixed-ip": entry.get("fixed_ip")}
                    for entry in nw_port_id_list]
            kwargs.update(nics=nics)
        if secgroup_name:
            kwargs.update(security_groups=[secgroup_name])

        try:
            nova = nova_client.Client(self.nova_version, auth_token=token,
                                      tenant_id=tenant_id,
                                      auth_url=self.identity_service)
            flavor = nova.flavors.find(name=flavor)
            instance = nova.servers.create(name, nova.images.get(image_id),
                                           flavor, **kwargs)
            data = instance.to_dict()
            return data['id']
        except Exception as ex:
            err = ("Failed to create instance under tenant"
                   " %s  %s" % (tenant_id, ex))
            LOG.error(err)
            raise Exception(err)


class NeutronClient(OpenstackApi):
    """ Neutron Client Api Driver. """

    def get_floating_ip(self, token, floatingip_id):
        """ Get floatingip details

        :param token: A scoped_token
        :param floatingip_id: Port UUID

        :return: floatingip details

        """
        try:
            neutron = neutron_client.Client(token=token,
                                            endpoint_url=self.network_service)
            return neutron.show_floatingip(floatingip_id)['floatingip']
        except Exception as ex:
            err = ("Failed to read floatingip from"
                   " Openstack Neutron service's response"
                   " KeyError :: %s" % (ex))
            LOG.error(err)
            raise Exception(err)

    def get_floating_ips(self, token, **filters):
        """ Get list of floatingips, associated with port if passed"""
        try:
            neutron = neutron_client.Client(token=token,
                                            endpoint_url=self.network_service)
            return neutron.list_floatingips(**filters)['floatingips']
        except Exception as ex:
            err = ("Failed to read floatingips from"
                   " Openstack Neutron service's response"
                   " KeyError :: %s" % (ex))
            LOG.error(err)
            raise Exception(err)

    def get_security_groups(self, token, tenant_id=None, filters=None):
        """ Get list of security groups"""
        try:
            neutron = neutron_client.Client(token=token,
                                            endpoint_url=self.network_service)
            filters = filters if filters is not None else {}
            return neutron.list_security_groups(**filters)['security_groups']
        except Exception as ex:
            err = ("Failed to get security groups from"
                   " Openstack Neutron service's response"
                   " KeyError :: %s" % (ex))
            LOG.error(err)
            raise Exception(err)

    def create_security_group(self, token, attrs=None):
        """ Create security group"""
        try:
            neutron = neutron_client.Client(token=token,
                                            endpoint_url=self.network_service)

            sg_info = {"security_group": attrs}
            return neutron.create_security_group(body=sg_info)[
                'security_group']
        except Exception as ex:
            err = ("Failed to get security groups from"
                   " Openstack Neutron service's response"
                   " KeyError :: %s" % (ex))
            LOG.error(err)
            raise Exception(err)

    def create_security_group_rule(self, token, attrs=None):
        """ Create security group rule"""
        try:
            neutron = neutron_client.Client(token=token,
                                            endpoint_url=self.network_service)

            # attrs={'direction': 'egress', 'protocol': 'TCP',
            # 'security_group_id': 'c90c7b29-f653-4c41-ae1a-0290dc64e020'}
            sg_rule_info = {"security_group_rule": attrs}
            return neutron.create_security_group_rule(
                body=sg_rule_info)['security_group_rule']
        except Exception as ex:
            err = ("Failed to get security groups from"
                   " Openstack Neutron service's response"
                   " KeyError :: %s" % (ex))
            LOG.error(err)
            # raise Exception(err)

    def get_ports(self, token, filters=None):
        """ List Ports

        :param token: A scoped_token
        :param filters: Parameters for list filter
        example for filter: ?tenant_id=%s&id=%s

        :return: Port List

        """
        try:
            neutron = neutron_client.Client(token=token,
                                            endpoint_url=self.network_service)
            ports = neutron.list_ports(**filters).get('ports', [])
            return ports
        except Exception as ex:
            err = ("Failed to read port list from"
                   " Openstack Neutron service's response"
                   " KeyError :: %s" % (ex))
            LOG.error(err)
            raise Exception(err)

    def get_port(self, token, port_id):
        """ Get port details

        :param token: A scoped_token
        :param port_id: Port UUID

        :return: Port details

        """
        try:
            neutron = neutron_client.Client(token=token,
                                            endpoint_url=self.network_service)
            return neutron.show_port(port_id)
        except Exception as ex:
            err = ("Failed to read port information"
                   " Exception :: %s" % (ex))
            LOG.error(err)
            raise Exception(err)

    def get_subnets(self, token, filters=None):
        """ List subnets

        :param token: A scoped_token
        :param filters: Parameters for list filter
        example for filter: ?tenant_id=%s&id=%s

        :return: Subnet List

        """
        try:
            neutron = neutron_client.Client(token=token,
                                            endpoint_url=self.network_service)
            subnets = neutron.list_subnets(**filters).get('subnets', [])
            return subnets
        except Exception as ex:
            err = ("Failed to read subnet list from"
                   " Openstack Neutron service's response"
                   " KeyError :: %s" % (ex))
            LOG.error(err)
            raise Exception(err)

    def get_subnet(self, token, subnet_id):
        """ Get subnet details
        :param token: A scoped_token
        :param subnet_id: Subnet UUID
        :return: Subnet details
        """
        try:
            neutron = neutron_client.Client(token=token,
                                            endpoint_url=self.network_service)
            return neutron.show_subnet(subnet_id)
        except Exception as ex:
            err = ("Failed to read subnet from"
                   " Openstack Neutron service's response"
                   " KeyError :: %s" % (ex))
            LOG.error(err)
            raise Exception(err)

    def delete_floatingip(self, token, floatingip_id):
        """ Delete the floatingip
        :param token: A scoped token
        :param tenant_id: Tenant UUID
        :param floatingip_id: Floatingip UUID
        """
        try:
            neutron = neutron_client.Client(token=token,
                                            endpoint_url=self.network_service)
            return neutron.delete_floatingip(floatingip_id)
        except Exception as ex:
            err = ("Failed to delete floatingip from"
                   " Openstack Neutron service's response"
                   " KeyError :: %s" % (ex))
            LOG.error(err)
            raise Exception(err)

    def update_port(self, token, port_id, **kwargs):
        """
        :param token:
        :param port_id:
        :param kwargs: name=<>, allowed_address_pairs={'ip_address': <>,
                       'mac_address': <>}
        :return:
        """
        try:
            neutron = neutron_client.Client(token=token,
                                            endpoint_url=self.network_service)
            port_info = dict(port={})
            port_info['port'].update(kwargs)
            return neutron.update_port(port_id, body=port_info)
        except Exception as ex:
            err = ("Failed to update port info"
                   " Error :: %s" % (ex))
            LOG.error(err)
            raise Exception(err)

    def get_floating_ips_for_ports(self, token, **kwargs):
        """
        :param self:
        :param token:
        :param kwargs:
        :return:
        """
        data = {'floatingips': []}
        try:
            neutron = neutron_client.Client(token=token,
                                            endpoint_url=self.network_service)
            data = neutron.list_floatingips(port_id=[kwargs[key]
                                                     for key in kwargs])
            return data
        except Exception as ex:
            raise Exception(ex)

    def _update_floatingip(self, token, floatingip_id, data):
        """ Update the floatingip
        :param token: A scoped token
        :param floatingip_id: Floatingip UUID
        :param data: data to update
        """
        try:
            neutron = neutron_client.Client(token=token,
                                            endpoint_url=self.network_service)
            return neutron.update_floatingip(floatingip_id, body=data)
        except Exception as ex:
            err = ("Failed to update floatingip from"
                   " Openstack Neutron service's response"
                   " KeyError :: %s" % (ex))
            LOG.error(err)
            raise Exception(err)

    def disassociate_floating_ip(self, token, floatingip_id):
        """
        :param self:
        :param token:
        :param floatingip_id:
        :return:
        """
        info = {
            "floatingip": {
                "port_id": None}
        }
        self._update_floatingip(token, floatingip_id, info)
        LOG.debug("Successfully disassociated floatingip %s",
                  floatingip_id)

    def associate_floating_ip(self, token, floatingip_id, port_id):
        """
        :param self:
        :param token:
        :param floatingip_id:
        :return:
        """

        info = {
            "floatingip": {
                "port_id": port_id}
        }

        self._update_floatingip(token, floatingip_id, info)
        LOG.debug("Successfully associated floatingip %s",
                  floatingip_id)

    def list_ports(self, token, port_ids=None, **kwargs):
        """
        :param token:
        :param port_ids:
        :param kwargs:
        :return:
        """
        try:
            neutron = neutron_client.Client(token=token,
                                            endpoint_url=self.network_service)
            port_ids = port_ids if port_ids is not None else []
            if port_ids:
                ports = neutron.list_ports(id=port_ids).get('ports', [])
            else:
                ports = neutron.list_ports(**kwargs)
            return ports
        except Exception as ex:
            err = ("Failed to list ports %s" % ex)
            LOG.error(err)
            raise Exception(err)

    def list_subnets(self, token, subnet_ids=None, **kwargs):
        """
        :param token:
        :param subnet_ids:
        :param kwargs:
        :return:
        """
        try:
            neutron = neutron_client.Client(token=token,
                                            endpoint_url=self.network_service)
            subnet_ids = subnet_ids if subnet_ids is not None else []
            subnets = neutron.list_subnets(id=subnet_ids).get('subnets', [])
            return subnets
        except Exception as ex:
            err = ("Failed to list subnets %s" % ex)
            LOG.error(err)
            raise Exception(err)

    def create_port(self, token, tenant_id, net_id, attrs=None):

        attr = {
            'port': {
                # 'tenant_id': tenant_id,
                'network_id': net_id
            }
        }

        if attrs:
            attr['port'].update(attrs)

        try:
            neutron = neutron_client.Client(token=token,
                                            endpoint_url=self.network_service)
            return neutron.create_port(body=attr)['port']
        except Exception as ex:
            raise Exception(_("Port creation failed in network: %(net)r "
                              "of tenant: %(tenant)r Error: %(error)s") %
                            {'net': net_id,
                             'tenant': tenant_id,
                             'error': ex})

    def delete_port(self, token, port_id):
        """
        :param token:
        :param port_id:
        :return:
        """
        try:
            neutron = neutron_client.Client(token=token,
                                            endpoint_url=self.network_service)
            return neutron.delete_port(port_id)
        except Exception as ex:
            err = ("Failed to delete port %s"
                   " Exception :: %s" % (port_id, ex))
            LOG.error(err)
            raise Exception(err)

    def get_networks(self, token, filters=None):
        """ List nets

        :param token: A scoped_token
        :param filters: Parameters for list filter
        example for filter: ?tenant_id=%s&id=%s

        :return: network List

        """
        try:
            neutron = neutron_client.Client(token=token,
                                            endpoint_url=self.network_service)
            nets = neutron.list_networks(**filters).get('networks', [])
            return nets
        except Exception as ex:
            err = ("Failed to read network list from"
                   " Openstack Neutron service's response"
                   " KeyError :: %s" % (ex))
            LOG.error(err)
            raise Exception(err)

    def delete_nw(self, token, net_id):
        """
        :param token:
        :param net_id:
        :return:
        """
        try:
            neutron = neutron_client.Client(token=token,
                                            endpoint_url=self.network_service)
            return neutron.delete_network(net_id)
        except Exception as ex:
            err = ('Failed to delete network %s . %s' % (net_id, str(ex)))
            LOG.error(err)

    def get_pools(self, token, filters=None):
        """ List Pools

        :param token: A scoped_token
        :param filters: Parameters for list filter
        example for filter: ?tenant_id=%s&id=%s

        :return: Pool List

        """
        try:
            neutron = neutron_client.Client(token=token,
                                            endpoint_url=self.network_service)
            pools = neutron.list_pools(**filters).get('pools', [])
            return pools
        except Exception as ex:
            err = ("Failed to read pool list from"
                   " Openstack Neutron service's response"
                   " KeyError :: %s" % (ex))
            LOG.error(err)
            raise Exception(err)

    def get_loadbalancers(self, token, filters=None):
        """ List Loadbalancers

        :param token: A scoped_token
        :param filters: Parameters for list filter
        example for filter: ?tenant_id=%s&id=%s

        :return: Loadbalancers List

        """
        try:
            neutron = neutron_client.Client(token=token,
                                            endpoint_url=self.network_service)
            loadbalancers = neutron.list_loadbalancers(**filters).get(
                'loadbalancers', [])
            return loadbalancers
        except Exception as ex:
            err = ("Failed to read pool list from"
                   " Openstack Neutron service's response"
                   " KeyError :: %s" % (ex))
            LOG.error(err)
            raise Exception(err)

    def get_vip(self, token, vip_id):
        """ Get vip details

        :param token: A scoped_token
        :param vip_id: Port UUID

        :return: VIP details

        """
        try:
            neutron = neutron_client.Client(token=token,
                                            endpoint_url=self.network_service)
            return neutron.show_vip(vip_id)
        except Exception as ex:
            err = ("Failed to read vip information"
                   " Exception :: %s" % (ex))
            LOG.error(err)
            raise Exception(err)

    def get_agents(self, token, filters=None):
        """ Get neutron agents

        :param token: A scoped_token
        :param filters: Parameters for list filter

        :return: neutron agents List

        """
        try:
            neutron = neutron_client.Client(token=token,
                                            endpoint_url=self.network_service)
            return neutron.list_agents(**filters).get('agents', [])
        except Exception as ex:
            err = ("Failed to read agents information"
                   " Exception :: %s" % (ex))
            LOG.error(err)
            raise Exception(err)


class GBPClient(OpenstackApi):
    """ GBP Client Api Driver. """

    def get_policy_target_groups(self, token, filters=None):
        """ List Policy Target Groups

        :param token: A scoped_token
        :param filters: Parameters for list filter
        example for filter: ?tenant_id=%s&id=%s

        :return: PTG List

        """
        try:
            gbp = gbp_client.Client(token=token,
                                    endpoint_url=self.network_service)
            return gbp.list_policy_target_groups(
                **filters)['policy_target_groups']
        except Exception as ex:
            err = ("Failed to read PTG list from"
                   " Openstack Neutron service's response."
                   " KeyError :: %s" % (ex))
            LOG.error(err)
            raise Exception(err)

    def get_policy_target_group(self, token, ptg_id, filters=None):
        """
        :param token: A scoped token
        :param ptg_id: PTG
        :param filters: Optional
        :return:
        """
        try:
            gbp = gbp_client.Client(token=token,
                                    endpoint_url=self.network_service)
            filters = filters if filters is not None else {}
            return gbp.show_policy_target_group(
                ptg_id, **filters)['policy_target_group']
        except Exception as ex:
            err = ("Failed to read PTG list from"
                   " Openstack Neutron service's response."
                   " KeyError :: %s" % (ex))
            LOG.error(err)
            raise Exception(err)

    def update_policy_target_group(self, token, ptg_id,
                                   policy_target_group_info):
        """ Updates a GBP Policy Target Group

        :param token: A scoped token
        :param ptg_id: PTG UUID
        :param policy_target_group_info: PTG info dict
        :return: PTG dict
        """
        try:
            gbp = gbp_client.Client(token=token,
                                    endpoint_url=self.network_service)
            return gbp.update_policy_target_group(
                ptg_id,
                body=policy_target_group_info)['policy_target_group']
        except Exception as ex:
            err = ("Failed to update policy target group. Error :: %s" % (ex))
            LOG.error(err)
            raise Exception(err)

    def create_policy_target(self, token, tenant_id,
                             policy_target_group_id, name, port_id=None,
                             description=''):
        """ Creates a GBP Policy Target

        :param token: A scoped token
        :param tenant_id: Tenant UUID
        :param policy_target_group_id: PTG UUID
        :param name: PT name
        :return: PT dict
        """
        policy_target_info = {
            "policy_target": {
                "policy_target_group_id": policy_target_group_id,
                "tenant_id": tenant_id,
            }
        }
        if name:
            policy_target_info['policy_target'].update({'name': name})
        if port_id:
            policy_target_info["policy_target"]["port_id"] = port_id
        if description:
            policy_target_info["policy_target"]["description"] = description

        try:
            gbp = gbp_client.Client(token=token,
                                    endpoint_url=self.network_service)
            return gbp.create_policy_target(
                body=policy_target_info)['policy_target']

        except Exception as ex:
            err = ("Failed to read policy target information from"
                   " Openstack Neutron service's response."
                   " KeyError :: %s" % (ex))
            LOG.error(err)
            raise Exception(err)

    def delete_policy_target(self, token, policy_target_id):
        """ Delete the GBP policy_target
        :param token: A scoped token
        :param policy_target_id: PT UUID
        """
        try:
            gbp = gbp_client.Client(token=token,
                                    endpoint_url=self.network_service)
            return gbp.delete_policy_target(policy_target_id)

        except Exception as ex:
            err = ("Failed to delete policy target information from"
                   " Openstack Neutron service's response."
                   " KeyError :: %s" % (ex))
            LOG.error(err)
            raise Exception(err)

    def delete_policy_target_group(self, token, policy_target_group_id):
        """ Delete the GBP policy_target group
        :param token: A scoped token
        :param policy_target_id: PTG UUID
        """
        try:
            gbp = gbp_client.Client(token=token,
                                    endpoint_url=self.network_service)
            return gbp.delete_policy_target_group(policy_target_group_id)
        except Exception as ex:
            err = ("Failed to delete policy target group from"
                   " Openstack."
                   " Error :: %s" % (ex))
            LOG.error(err)
            raise Exception(err)

    def update_policy_target(self, token, policy_target_id, updated_pt):
        """ Update the Policy Target
        :param token: A scoped token
        :param policy_target_id: PT UUID
        :param updated_pt: New PT dict
        {\"policy_target\": {\"description\": \"test123\"}}
        """

        policy_target_info = {
            "policy_target": updated_pt
        }

        try:
            gbp = gbp_client.Client(token=token,
                                    endpoint_url=self.network_service)
            return gbp.update_policy_target(
                policy_target_id, body=policy_target_info)['policy_target']
        except Exception as ex:
            err = ("Failed to read updated PT information"
                   ". PT  %s."
                   " KeyError :: %s" % (policy_target_id, ex))
            LOG.error(err)
            raise Exception(err)

    def create_policy_target_group(self, token, tenant_id, name,
                                   l2_policy_id=None, ext_data=None):
        """ Creates a GBP Policy Target Group

        :param ext_data:
        :param token: A scoped token
        :param tenant_id: Tenant UUID
        :param name: PTG name
        :return: PTG dict
        """

        policy_target_group_info = {
            "policy_target_group": {
                "tenant_id": tenant_id,
                "name": name,
            }
        }
        ext_data = ext_data or {}
        if l2_policy_id:
            policy_target_group_info["policy_target_group"].update(
                {"l2_policy_id": l2_policy_id})
        policy_target_group_info['policy_target_group'].update(ext_data)

        try:
            gbp = gbp_client.Client(token=token,
                                    endpoint_url=self.network_service)
            return gbp.create_policy_target_group(
                body=policy_target_group_info)['policy_target_group']
        except Exception as ex:
            err = ("Failed to create policy target group. %s"
                   " Error :: %s" % (policy_target_group_info, ex))
            LOG.error(err)
            raise Exception(err)

    def create_l2_policy(self, token, tenant_id, name, l3_policy_id=None,
                         description=''):

        l2_policy_info = {
            "l2_policy": {
                "tenant_id": tenant_id,
                "name": name
            }
        }
        if l3_policy_id:
            l2_policy_info["l2_policy"].update({'l3_policy_id': l3_policy_id})
        if description:
            l2_policy_info["description"].update({'description': description})

        try:
            gbp = gbp_client.Client(token=token,
                                    endpoint_url=self.network_service)
            return gbp.create_l2_policy(body=l2_policy_info)['l2_policy']
        except Exception as ex:
            err = ("Failed to create l2 policy under tenant"
                   " %s. Error :: %s" % (tenant_id, ex))
            LOG.error(err)
            raise Exception(err)

    def delete_l2_policy(self, token, l2policy_id):
        """
        :param token:
        :param l2policy_id:
        :return:
        """
        try:
            gbp = gbp_client.Client(token=token,
                                    endpoint_url=self.network_service)
            return gbp.delete_l2_policy(l2policy_id)
        except Exception as ex:
            err = ("Failed to delete l2 policy %s. Reason %s" %
                   (l2policy_id, ex))
            LOG.error(err)
            raise Exception(err)

    # NOTE: The plural form in the function name is needed in that way
    # to construct the function generically
    def get_l2_policys(self, token, filters=None):
        """ List L2 policies

        :param token: A scoped_token
        :param filters: Parameters for list filter
        example for filter: {}

        :return: L2 policies List

        """
        try:
            gbp = gbp_client.Client(token=token,
                                    endpoint_url=self.network_service)
            filters = filters if filters is not None else {}
            return gbp.list_l2_policies(**filters)['l2_policies']
        except Exception as ex:
            err = ("Failed to list l2 policies. Reason %s" % ex)
            LOG.error(err)
            raise Exception(err)

    def get_l2_policy(self, token, policy_id, filters=None):
        """ List L2 policies

        :param token: A scoped_token
        :param policy_id: l2 policy id
        :param filters: Parameters for list filter
        example for filter: {}

        :return: L2 policies List

        """
        try:
            gbp = gbp_client.Client(token=token,
                                    endpoint_url=self.network_service)
            filters = filters if filters is not None else {}
            return gbp.show_l2_policy(
                policy_id, **filters)['l2_policy']
        except Exception as ex:
            err = ("Failed to read l2 policy list from"
                   " Openstack Neutron service's response."
                   " KeyError :: %s" % (ex))
            LOG.error(err)
            raise Exception(err)

    def create_network_service_policy(self, token,
                                      network_service_policy_info):

        try:
            gbp = gbp_client.Client(token=token,
                                    endpoint_url=self.network_service)
            return gbp.create_network_service_policy(
                body=network_service_policy_info)['network_service_policy']
        except Exception as ex:
            err = ("Failed to create network service policy "
                   "Error :: %s" % (ex))
            LOG.error(err)
            raise Exception(err)

    def get_network_service_policies(self, token, filters=None):
        """ List network service policies

        :param token: A scoped_token
        :param filters: Parameters for list filter
        example for filter: {}

        :return: network service policy List

        """
        try:
            gbp = gbp_client.Client(token=token,
                                    endpoint_url=self.network_service)
            filters = filters if filters is not None else {}
            return gbp.list_network_service_policies(**filters)[
                'network_service_policies']
        except Exception as ex:
            err = ("Failed to list network service policies. Reason %s" % ex)
            LOG.error(err)
            raise Exception(err)

    def get_external_policies(self, token, filters=None):
        """ List external policies

        :param token: A scoped_token
        :param filters: Parameters for list filter
        example for filter: {}

        :return: external policy List

        """
        try:
            gbp = gbp_client.Client(token=token,
                                    endpoint_url=self.network_service)
            filters = filters if filters is not None else {}
            return gbp.list_external_policies(**filters)['external_policies']
        except Exception as ex:
            err = ("Failed to list external policies. Reason %s" % ex)
            LOG.error(err)
            raise Exception(err)

    def get_policy_rule_sets(self, token, filters=None):
        """ List policy rule sets

        :param token: A scoped_token
        :param filters: Parameters for list filter
        example for filter: {}

        :return: policy rule set List

        """
        try:
            gbp = gbp_client.Client(token=token,
                                    endpoint_url=self.network_service)
            filters = filters if filters is not None else {}
            return gbp.list_policy_rule_sets(**filters)['policy_rule_sets']
        except Exception as ex:
            err = ("Failed to list policy rule sets. Reason %s" % ex)
            LOG.error(err)
            raise Exception(err)

    def get_policy_actions(self, token, filters=None):
        """ List policy actions

        :param token: A scoped_token
        :param filters: Parameters for list filter
        example for filter: {}

        :return: policy actions List

        """
        try:
            gbp = gbp_client.Client(token=token,
                                    endpoint_url=self.network_service)
            filters = filters if filters is not None else {}
            return gbp.list_policy_actions(**filters)['policy_actions']
        except Exception as ex:
            err = ("Failed to list policy actions. Reason %s" % ex)
            LOG.error(err)
            raise Exception(err)

    def get_policy_rules(self, token, filters=None):
        """ List policy rules

        :param token: A scoped_token
        :param filters: Parameters for list filter
        example for filter: {}

        :return: policy rules List

        """
        try:
            gbp = gbp_client.Client(token=token,
                                    endpoint_url=self.network_service)
            filters = filters if filters is not None else {}
            return gbp.list_policy_rules(**filters)['policy_rules']
        except Exception as ex:
            err = ("Failed to list policy rules. Reason %s" % ex)
            LOG.error(err)
            raise Exception(err)

    def create_l3_policy(self, token, l3_policy_info):  # tenant_id, name):

        try:
            gbp = gbp_client.Client(token=token,
                                    endpoint_url=self.network_service)
            return gbp.create_l3_policy(body=l3_policy_info)['l3_policy']
        except Exception as ex:
            err = ("Failed to create l3 policy under tenant"
                   " %s. Error :: %s"
                   % (l3_policy_info['l3_policy']['tenant_id'], ex))
            LOG.error(err)
            raise Exception(err)

    def get_l3_policy(self, token, policy_id, filters=None):
        """ List L3 policies

        :param token: A scoped_token
        :param filters: Parameters for list filter
        example for filter: {}

        :return: L3 policies List

        """
        try:
            gbp = gbp_client.Client(token=token,
                                    endpoint_url=self.network_service)
            filters = filters if filters is not None else {}
            return gbp.show_l3_policy(
                policy_id, **filters)['l3_policy']
        except Exception as ex:
            err = ("Failed to read l3 policy list from"
                   " Openstack Neutron service's response."
                   " KeyError :: %s" % (ex))
            LOG.error(err)
            raise Exception(err)

    def get_l3_policies(self, token, filters=None):
        """ List L3 policies

        :param token: A scoped_token
        :param filters: Parameters for list filter
        example for filter: {}

        :return: L2 policies List

        """
        try:
            gbp = gbp_client.Client(token=token,
                                    endpoint_url=self.network_service)
            filters = filters if filters is not None else {}
            return gbp.list_l3_policies(**filters)['l3_policies']
        except Exception as ex:
            err = ("Failed to list l3 policies. Reason %s" % ex)
            LOG.error(err)
            raise Exception(err)

    def get_policy_targets(self, token, filters=None):
        """ List Policy Targets

        :param token: A scoped_token
        :param filters: Parameters for list filter
        example for filter: {}

        :return: PT List

        """
        try:
            gbp = gbp_client.Client(token=token,
                                    endpoint_url=self.network_service)
            filters = filters if filters is not None else {}
            return gbp.list_policy_targets(**filters)['policy_targets']
        except Exception as ex:
            err = ("Failed to read PT list."
                   " Error :: %s" % (ex))
            LOG.error(err)
            raise Exception(err)

    def list_pt(self, token, filters=None):
        filters = filters if filters is not None else {}
        return self.get_policy_targets(token, filters=filters)

    def get_policy_target(self, token, pt_id, filters=None):
        try:
            gbp = gbp_client.Client(token=token,
                                    endpoint_url=self.network_service)
            filters = filters if filters is not None else {}
            return gbp.show_policy_target(pt_id,
                                          **filters)['policy_target']
        except Exception as ex:
            err = ("Failed to read PT information"
                   ". PT  %s."
                   " Error :: %s" % (pt_id, ex))
            LOG.error(err)
            raise Exception(err)

    def get_service_profile(self, token, service_profile_id):
        gbp = gbp_client.Client(token=token,
                                endpoint_url=self.network_service)
        return gbp.show_service_profile(service_profile_id)['service_profile']

    def get_servicechain_node(self, token, node_id):
        gbp = gbp_client.Client(token=token,
                                endpoint_url=self.network_service)
        return gbp.show_servicechain_node(node_id)['servicechain_node']

    def get_servicechain_instance(self, token, instance_id):
        gbp = gbp_client.Client(token=token,
                                endpoint_url=self.network_service)
        return gbp.show_servicechain_instance(instance_id)[
            'servicechain_instance']



# Copyright (c) 2014 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.db import utils as db_utils
from oslo_db.sqlalchemy import utils as sa_utils
import six
import weakref


# TODO(ashu): Below class need to extend neutron's CommonDbMixin.
class CommonDbMixin(object):
    """Common methods used in core and service plugins."""
    # Plugins, mixin classes implementing extension will register
    # hooks into the dict below for "augmenting" the "core way" of
    # building a query for retrieving objects from a model class.
    # To this aim, the register_model_query_hook and unregister_query_hook
    # from this class should be invoked
    _model_query_hooks = {}

    # This dictionary will store methods for extending attributes of
    # api resources. Mixins can use this dict for adding their own methods
    _dict_extend_functions = {}

    @classmethod
    def register_model_query_hook(cls, model, name, query_hook, filter_hook,
                                  result_filters=None):
        """Register a hook to be invoked when a query is executed.

        Add the hooks to the _model_query_hooks dict. Models are the keys
        of this dict, whereas the value is another dict mapping hook names to
        callables performing the hook.
        Each hook has a "query" component, used to build the query expression
        and a "filter" component, which is used to build the filter expression.

        Query hooks take as input the query being built and return a
        transformed query expression.

        Filter hooks take as input the filter expression being built and return
        a transformed filter expression
        """
        model_hooks = cls._model_query_hooks.get(model)
        if not model_hooks:
            # add key to dict
            model_hooks = {}
            cls._model_query_hooks[model] = model_hooks
        model_hooks[name] = {'query': query_hook, 'filter': filter_hook,
                             'result_filters': result_filters}

    @property
    def safe_reference(self):
        """Return a weakref to the instance.

        Minimize the potential for the instance persisting
        unnecessarily in memory by returning a weakref proxy that
        won't prevent deallocation.
        """
        return weakref.proxy(self)

    def _model_query(self, session, model, is_admin=False):
        query = session.query(model)
        # define basic filter condition for model query
        query_filter = None
        # Execute query hooks registered from mixins and plugins
        query_hooks = self._model_query_hooks.get(model, {})
        for _name, hooks in six.iteritems(query_hooks):
            query_hook = hooks.get('query')
            if isinstance(query_hook, six.string_types):
                query_hook = getattr(self, query_hook, None)
            if query_hook:
                query = query_hook(model, query)

            filter_hook = hooks.get('filter')
            if isinstance(filter_hook, six.string_types):
                filter_hook = getattr(self, filter_hook, None)
            if filter_hook:
                query_filter = filter_hook(model, query_filter)

        if query_filter is not None:
            query = query.filter(query_filter)
        return query

    def _fields(self, resource, fields):
        if fields:
            return dict(((key, item) for key, item in resource.items()
                         if key in fields))
        return resource

    def _get_tenant_id_for_create(self, resource):
        return resource['tenant_id']

    def _get_by_id(self, session, model, id):
        query = self._model_query(session, model)
        return query.filter(model.id == id).one()

    def _apply_filters_to_query(self, query, model, filters):
        if filters:
            for key, value in six.iteritems(filters):
                column = getattr(model, key, None)
                if column:
                    query = query.filter(column.in_(value))

            query_hooks = self._model_query_hooks.get(model, {})
            for _name, hooks in six.iteritems(query_hooks):
                result_filter = hooks.get('result_filters', None)
                if isinstance(result_filter, six.string_types):
                    result_filter = getattr(self, result_filter, None)

                if result_filter:
                    query = result_filter(query, filters)
        return query

    def _apply_dict_extend_functions(self, resource_type,
                                     response, db_object):
        for func in self._dict_extend_functions.get(
                resource_type, []):
            args = (response, db_object)
            if isinstance(func, six.string_types):
                func = getattr(self, func, None)
            else:
                # must call unbound method - use self as 1st argument
                args = (self,) + args
            if func:
                func(*args)

    def _get_collection_query(self, session, model, filters=None,
                              sorts=None, limit=None, marker_obj=None,
                              page_reverse=False):
        collection = self._model_query(session, model)
        collection = self._apply_filters_to_query(collection, model, filters)
        if sorts:
            sort_keys = db_utils.get_and_validate_sort_keys(sorts, model)
            sort_dirs = db_utils.get_sort_dirs(sorts, page_reverse)
            collection = sa_utils.paginate_query(
                collection, model, limit,
                marker=marker_obj,
                sort_keys=sort_keys,
                sort_dirs=sort_dirs)

        return collection

    def _get_collection(self, session, model, dict_func, filters=None,
                        fields=None, sorts=None, limit=None, marker_obj=None,
                        page_reverse=False):
        query = self._get_collection_query(session, model, filters=filters,
                                           sorts=sorts,
                                           limit=limit,
                                           marker_obj=marker_obj,
                                           page_reverse=page_reverse)
        items = [dict_func(c, fields) for c in query]
        if limit and page_reverse:
            items.reverse()
        return items

    def _get_collection_count(self, model, filters=None):
        return self._get_collection_query(model, filters).count()

    def _get_marker_obj(self, resource, limit, marker):
        if limit and marker:
            return getattr(self, '_get_%s' % resource)(marker)
        return None

    def _filter_non_model_columns(self, data, model):
        """Filter non model columns

        Remove all the attributes from data which are not columns of
        the model passed as second parameter.
        """
        columns = [c.name for c in model.__table__.columns]
        return dict((k, v) for (k, v) in
                    six.iteritems(data) if k in columns)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib import exceptions as n_exc
from oslo_config import cfg
from oslo_log import log as logging
import stevedore

from gbpservice._i18n import _LI
from gbpservice.neutron.services.servicechain.plugins.ncp import config  # noqa
from gbpservice.neutron.services.servicechain.plugins.ncp import model

LOG = logging.getLogger(__name__)


class NodeDriverManager(stevedore.named.NamedExtensionManager):
    """Route servicechain APIs to servicechain node drivers.

    """

    def __init__(self):
        # Registered node drivers, keyed by name.
        self.drivers = {}
        # Ordered list of node drivers.
        self.ordered_drivers = []
        names = cfg.CONF.node_composition_plugin.node_drivers
        LOG.info(_LI("Configured service chain node driver names: %s"), names)
        super(NodeDriverManager,
              self).__init__(
                  'gbpservice.neutron.servicechain.ncp_drivers', names,
                  invoke_on_load=True, name_order=True)
        LOG.info(_LI(
            "Loaded service chain node driver names: %s"), self.names())
        self._register_drivers()

    def _register_drivers(self):
        """Register all service chain node drivers."""
        for ext in self:
            self.drivers[ext.name] = ext
            self.ordered_drivers.append(ext)
        LOG.info(_LI("Registered service chain node drivers: %s"),
                 [driver.name for driver in self.ordered_drivers])

    def initialize(self):
        """Initialize all the service chain node drivers."""
        self.native_bulk_support = True
        for driver in self.ordered_drivers:
            LOG.info(_LI("Initializing service chain node drivers '%s'"),
                     driver.name)
            driver.obj.initialize(driver.name)
            self.native_bulk_support &= getattr(driver.obj,
                                                'native_bulk_support', True)

    def schedule_deploy(self, context):
        """Schedule Node Driver for Node creation.

        Given a NodeContext, this method returns the driver capable of creating
        the specific node.
        """
        for driver in self.ordered_drivers:
            try:
                driver.obj.validate_create(context)
                if not model.get_node_owner(context):
                    model.set_node_owner(context, driver.obj.name)
                return driver.obj
            except n_exc.NeutronException as e:
                LOG.warning(e.message)

    def schedule_destroy(self, context):
        """Schedule Node Driver for Node disruption.

        Given a NodeContext, this method returns the driver capable of
        destroying the specific node.
        """
        driver = self.get_owning_driver(context)
        if driver:
            model.unset_node_owner(context)
        return driver

    def schedule_update(self, context):
        """Schedule Node Driver for Node Update.

        Given a NodeContext, this method returns the driver capable of updating
        the specific node.
        """
        driver = self.get_owning_driver(context)
        if driver:
            driver.validate_update(context)
        return driver

    def schedule_get(self, context):
        """Schedule Node Driver to get Node details.

        Given a NodeContext, this method returns the driver capable of
        getting details the specific node.
        """
        driver = self.get_owning_driver(context)
        return driver

    def clear_node_owner(self, context):
        """Remove Node Driver ownership set for a Node

        Given a NodeContext, this method removes the Node owner mapping in DB.
        This method is used when we want to perform a disruptive chain update
        by deleting and recreating the Node instances
        """
        model.unset_node_owner(context)

    def get_owning_driver(self, context):
        owner = model.get_node_owner(context)
        if owner:
            driver = self.drivers.get(owner[0].driver_name)
            return driver.obj if driver else None



# Copyright (c) 2017 Cisco Systems Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# Note: This module should be treated as legacy and should not be extended to
# add any new data migrations. New data migrations should be added
# directly to the alembic migration script along with the table definitions
# that are being referenced.
# For reference see how its done here:
# https://github.com/openstack/neutron/blob/
# 625de54de3936b0da8760c3da76d2d315d05f94e/neutron/db/migration/
# alembic_migrations/versions/newton/contract/
# 3b935b28e7a0_migrate_to_pluggable_ipam.py

import netaddr

from aim.aim_lib import nat_strategy
from aim import aim_manager
from aim.api import resource as aim_resource
from aim import context as aim_context
from aim import utils as aim_utils
from alembic import util as alembic_util
from neutron.db.models import address_scope as as_db
from neutron.db.models import securitygroup as sg_models
from neutron.db.migration.cli import *  # noqa
from neutron.db import models_v2
from neutron.db import segments_db  # noqa
from neutron_lib.db import model_base
import sqlalchemy as sa
from sqlalchemy.orm import lazyload

from gbpservice.neutron.extensions import cisco_apic as ext
from gbpservice.neutron.plugins.ml2plus.drivers.apic_aim import apic_mapper


# The following definitions have been taken from commit:
# 9b4b7276ad8a0f181c9be12ba5a0192432aa5027
# and is frozen for the data migration script that was included
# in this module. It should not be changed in this module.
NetworkExtensionDb = sa.Table(
        'apic_aim_network_extensions', sa.MetaData(),
        sa.Column('network_id', sa.String(36), nullable=False),
        sa.Column('external_network_dn', sa.String(1024)),
        sa.Column('nat_type', sa.Enum('distributed', 'edge', '')))


NetworkExtensionCidrDb = sa.Table(
        'apic_aim_network_external_cidrs', sa.MetaData(),
        sa.Column('network_id', sa.String(36), nullable=False),
        sa.Column('cidr', sa.String(64), nullable=False))


AddressScopeMapping = sa.Table(
        'apic_aim_address_scope_mappings', sa.MetaData(),
        sa.Column('scope_id', sa.String(36)),
        sa.Column('vrf_name', sa.String(64)),
        sa.Column('vrf_tenant_name', sa.String(64)),
        sa.Column('vrf_owned', sa.Boolean, nullable=False))


# The following definition has been taken from commit:
# f8b41855acbbb7e59a0bab439445c198fc6aa146
# and is frozen for the data migration script that was included
# in this module. It should not be changed in this module.
NetworkMapping = sa.Table(
        'apic_aim_network_mappings', sa.MetaData(),
        sa.Column('network_id', sa.String(36), nullable=False),
        sa.Column('bd_name', sa.String(64), nullable=True),
        sa.Column('bd_tenant_name', sa.String(64), nullable=True),
        sa.Column('epg_name', sa.String(64), nullable=True),
        sa.Column('epg_tenant_name', sa.String(64), nullable=True),
        sa.Column('epg_app_profile_name', sa.String(64), nullable=True),
        sa.Column('vrf_name', sa.String(64), nullable=True),
        sa.Column('vrf_tenant_name', sa.String(64), nullable=True))


class DefunctAddressScopeExtensionDb(model_base.BASEV2):
    # REVISIT: This DB model class is used only for the
    # apic_aim_persist data migration, after which this table is
    # dropped.

    __tablename__ = 'apic_aim_addr_scope_extensions'

    address_scope_id = sa.Column(
        sa.String(36), sa.ForeignKey('address_scopes.id', ondelete="CASCADE"),
        primary_key=True)
    vrf_dn = sa.Column(sa.String(1024))


def _add_address_scope_mapping(session, scope_id, vrf, vrf_owned=True):
    session.execute(AddressScopeMapping.insert().values(
        scope_id=scope_id, vrf_name=vrf.name, vrf_tenant_name=vrf.tenant_name,
        vrf_owned=vrf_owned))


def _add_network_mapping(session, network_id, bd, epg, vrf, ext_net=None):
    if not ext_net:
        session.execute(NetworkMapping.insert().values(
            network_id=network_id, bd_name=bd.name,
            bd_tenant_name=bd.tenant_name, epg_name=epg.name,
            epg_app_profile_name=epg.app_profile_name,
            epg_tenant_name=epg.tenant_name, vrf_name=vrf.name,
            vrf_tenant_name=vrf.tenant_name))


def do_apic_aim_persist_migration(session):
    alembic_util.msg(
        "Starting data migration for apic_aim mechanism driver persistence.")

    aim = aim_manager.AimManager()
    aim_ctx = aim_context.AimContext(session)
    mapper = apic_mapper.APICNameMapper()

    with session.begin(subtransactions=True):
        # Migrate address scopes.
        scope_dbs = (session.query(as_db.AddressScope)
                     .options(lazyload('*')).all())
        for scope_db in scope_dbs:
            alembic_util.msg("Migrating address scope: %s" % scope_db)
            vrf = None
            ext_db = (session.query(DefunctAddressScopeExtensionDb).
                      filter_by(address_scope_id=scope_db.id).
                      one_or_none())
            if ext_db:
                # It has a pre-existing VRF.
                vrf = aim_resource.VRF.from_dn(ext_db.vrf_dn)
                # REVISIT: Get VRF to verify it exists?
                vrf_owned = False
            if not vrf:
                # It does not have a pre-existing VRF.
                aname = mapper.address_scope(session, scope_db.id)
                vrfs = aim.find(
                    aim_ctx, aim_resource.VRF,
                    name=aname)
                if vrfs:
                    vrf = vrfs[0]
                    vrf_owned = True
            if vrf:
                _add_address_scope_mapping(
                    session, scope_db.id, vrf, vrf_owned)
            else:
                alembic_util.warn(
                    "No AIM VRF found for address scope: %s" % scope_db)

        # Migrate networks.
        net_dbs = (session.query(models_v2.Network)
                   .options(lazyload('*')).all())
        for net_db in net_dbs:
            alembic_util.msg("Migrating network: %s" % net_db)
            bd = None
            epg = None
            vrf = None
            ext_db = (session.query(NetworkExtensionDb).
                      filter_by(network_id=net_db.id).
                      one_or_none())
            if ext_db and ext_db.external_network_dn:
                # Its a managed external network.
                ext_net = aim_resource.ExternalNetwork.from_dn(
                    ext_db.external_network_dn)
                # REVISIT: Get ExternalNetwork to verify it exists?
                l3out = aim_resource.L3Outside(
                    tenant_name=ext_net.tenant_name,
                    name=ext_net.l3out_name)
                if ext_db.nat_type == '':
                    ns_cls = nat_strategy.NoNatStrategy
                elif ext_db.nat_type == 'edge':
                    ns_cls = nat_strategy.EdgeNatStrategy
                else:
                    ns_cls = nat_strategy.DistributedNatStrategy
                ns = ns_cls(aim)
                ns.app_profile_name = 'OpenStack'
                for resource in ns.get_l3outside_resources(aim_ctx, l3out):
                    if isinstance(resource, aim_resource.BridgeDomain):
                        bd = resource
                    elif isinstance(resource, aim_resource.EndpointGroup):
                        epg = resource
                    elif isinstance(resource, aim_resource.VRF):
                        vrf = resource
            if not bd:
                # It must be a normal network.
                aname = mapper.network(session, net_db.id)
                bds = aim.find(
                    aim_ctx, aim_resource.BridgeDomain,
                    name=aname)
                if bds:
                    bd = bds[0]
                epgs = aim.find(
                    aim_ctx, aim_resource.EndpointGroup,
                    name=aname)
                if epgs:
                    epg = epgs[0]
                if bd:
                    vrfs = (
                        aim.find(
                            aim_ctx, aim_resource.VRF,
                            tenant_name=bd.tenant_name,
                            name=bd.vrf_name) or
                        aim.find(
                            aim_ctx, aim_resource.VRF,
                            tenant_name='common',
                            name=bd.vrf_name))
                    if vrfs:
                        vrf = vrfs[0]
            if bd and epg and vrf:
                _add_network_mapping(session, net_db.id, bd, epg, vrf)
            elif not net_db.external:
                alembic_util.warn(
                    "AIM BD, EPG or VRF not found for network: %s" % net_db)

    alembic_util.msg(
        "Finished data migration for apic_aim mechanism driver persistence.")


def _get_network_extn_db(session, network_id):
    netres = (session.query(NetworkExtensionDb).filter_by(
              network_id=network_id).first())

    if netres:
        _, ext_net_dn, nat_type = netres
        db_cidrs = (session.query(NetworkExtensionCidrDb).filter_by(
                    network_id=network_id).all())
        result = {}
        if ext_net_dn is not None:
            result[ext.EXTERNAL_NETWORK] = ext_net_dn
        if nat_type is not None:
            result[ext.NAT_TYPE] = nat_type
        if result.get(ext.EXTERNAL_NETWORK):
            result[ext.EXTERNAL_CIDRS] = [c for _, c in db_cidrs]
        return result


def do_ap_name_change(session, conf=None):
    alembic_util.msg("Starting data migration for apic_aim ap name change.")
    cfg = conf or CONF
    aim = aim_manager.AimManager()
    aim_ctx = aim_context.AimContext(session)
    system_id = cfg.apic_system_id
    alembic_util.msg("APIC System ID: %s" % system_id)
    with session.begin(subtransactions=True):
        net_dbs = session.query(models_v2.Network).options(lazyload('*')).all()
        for net_db in net_dbs:
            ext_db = _get_network_extn_db(session, net_db.id)
            if ext_db and ext_db.get(ext.EXTERNAL_NETWORK):
                alembic_util.msg("Migrating external network: %s" % net_db)
                # Its a managed external network.
                ext_net = aim_resource.ExternalNetwork.from_dn(
                    ext_db[ext.EXTERNAL_NETWORK])
                ext_net = aim.get(aim_ctx, ext_net)
                l3out = aim_resource.L3Outside(tenant_name=ext_net.tenant_name,
                                               name=ext_net.l3out_name)
                if ext_db[ext.NAT_TYPE] == '':
                    ns_cls = nat_strategy.NoNatStrategy
                elif ext_db[ext.NAT_TYPE] == 'edge':
                    ns_cls = nat_strategy.EdgeNatStrategy
                else:
                    ns_cls = nat_strategy.DistributedNatStrategy
                clone_ext_nets = {}
                ns = ns_cls(aim)
                ns.app_profile_name = 'OpenStack'
                ns.common_scope = None
                # Start Cleanup
                if not isinstance(ns, nat_strategy.NoNatStrategy):
                    l3out_clones = ns.db.get_clones(aim_ctx, l3out)
                    # Retrieve External Networks
                    for l3out_clone in l3out_clones:
                        for extc in aim.find(
                                aim_ctx, aim_resource.ExternalNetwork,
                                tenant_name=l3out_clone[0],
                                l3out_name=l3out_clone[1]):
                            clone_ext_nets[(l3out.tenant_name,
                                            l3out.name,
                                            extc.name)] = extc
                vrfs = ns.read_vrfs(aim_ctx, ext_net)
                session.execute(NetworkMapping.delete().where(
                    NetworkMapping.c.network_id == net_db.id))
                for vrf in vrfs:
                    ns.disconnect_vrf(aim_ctx, ext_net, vrf)
                ns.delete_external_network(aim_ctx, ext_net)
                ns.delete_l3outside(aim_ctx, l3out)
                # Recreate
                ns.common_scope = system_id
                ns.create_l3outside(aim_ctx, l3out)
                ns.create_external_network(aim_ctx, ext_net)
                ns.update_external_cidrs(aim_ctx, ext_net,
                                         ext_db[ext.EXTERNAL_CIDRS])
                for subnet in net_db.subnets:
                    aim_subnet = aim_resource.Subnet.to_gw_ip_mask(
                        subnet.gateway_ip, int(subnet.cidr.split('/')[1]))
                    ns.create_subnet(aim_ctx, l3out, aim_subnet)
                for resource in ns.get_l3outside_resources(aim_ctx, l3out):
                    if isinstance(resource, aim_resource.BridgeDomain):
                        bd = resource
                    elif isinstance(resource, aim_resource.EndpointGroup):
                        epg = resource
                    elif isinstance(resource, aim_resource.VRF):
                        vrf = resource
                _add_network_mapping(session, net_db.id, bd, epg, vrf)
                eid = (ext_net.tenant_name, ext_net.l3out_name, ext_net.name)
                for vrf in vrfs:
                    if eid in clone_ext_nets:
                        ext_net.provided_contract_names = clone_ext_nets[
                            eid].provided_contract_names
                        ext_net.consumed_contract_names = clone_ext_nets[
                            eid].consumed_contract_names
                    ns.connect_vrf(aim_ctx, ext_net, vrf)


def do_apic_aim_security_group_migration(session):
    alembic_util.msg(
        "Starting data migration for SGs and its rules.")

    aim = aim_manager.AimManager()
    aim_ctx = aim_context.AimContext(session)
    mapper = apic_mapper.APICNameMapper()
    with session.begin(subtransactions=True):
        # Migrate SG.
        sg_dbs = (session.query(sg_models.SecurityGroup).
                  options(lazyload('*')).all())
        for sg_db in sg_dbs:
            alembic_util.msg("Migrating SG: %s" % sg_db)
            tenant_aname = mapper.project(session, sg_db['tenant_id'])
            sg_aim = aim_resource.SecurityGroup(
                tenant_name=tenant_aname, name=sg_db['id'],
                display_name=aim_utils.sanitize_display_name(sg_db['name']))
            aim.create(aim_ctx, sg_aim, overwrite=True)
            # Always create this default subject
            sg_subject = aim_resource.SecurityGroupSubject(
                tenant_name=tenant_aname,
                security_group_name=sg_db['id'], name='default')
            aim.create(aim_ctx, sg_subject, overwrite=True)

        # Migrate SG rules.
        sg_rule_dbs = (session.query(sg_models.SecurityGroupRule).
                       options(lazyload('*')).all())
        for sg_rule_db in sg_rule_dbs:
            tenant_aname = mapper.project(session, sg_rule_db['tenant_id'])
            if sg_rule_db.get('remote_group_id'):
                ip_version = 0
                if sg_rule_db['ethertype'] == 'IPv4':
                    ip_version = 4
                elif sg_rule_db['ethertype'] == 'IPv6':
                    ip_version = 6
                remote_ips = []
                sg_ports = (session.query(models_v2.Port).
                            join(sg_models.SecurityGroupPortBinding,
                                 sg_models.SecurityGroupPortBinding.port_id ==
                                 models_v2.Port.id).
                            filter(sg_models.SecurityGroupPortBinding.
                                   security_group_id ==
                                   sg_rule_db['remote_group_id']).
                            options(lazyload('*')).all())
                for sg_port in sg_ports:
                    for fixed_ip in sg_port['fixed_ips']:
                        if ip_version == netaddr.IPAddress(
                                fixed_ip['ip_address']).version:
                            remote_ips.append(fixed_ip['ip_address'])
            else:
                remote_ips = ([sg_rule_db['remote_ip_prefix']]
                              if sg_rule_db['remote_ip_prefix'] else '')
            sg_rule_aim = aim_resource.SecurityGroupRule(
                tenant_name=tenant_aname,
                security_group_name=sg_rule_db['security_group_id'],
                security_group_subject_name='default',
                name=sg_rule_db['id'],
                direction=sg_rule_db['direction'],
                ethertype=sg_rule_db['ethertype'].lower(),
                ip_protocol=(sg_rule_db['protocol'] if sg_rule_db['protocol']
                             else 'unspecified'),
                remote_ips=remote_ips,
                from_port=(sg_rule_db['port_range_min']
                           if sg_rule_db['port_range_min'] else 'unspecified'),
                to_port=(sg_rule_db['port_range_max']
                         if sg_rule_db['port_range_max'] else 'unspecified'))
            aim.create(aim_ctx, sg_rule_aim, overwrite=True)

    alembic_util.msg(
        "Finished data migration for SGs and its rules.")



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy

from gbpservice.nfp.common import constants as const
from gbpservice.nfp.core import log as nfp_logging

LOG = nfp_logging.getLogger(__name__)


class DataParser(object):
    ''' A library to parse device and service configuration and
    transform them into a dictionary of key-value pairs

    '''

    def __init__(self):
        pass

    def parse_data(self, resource, data):
        ''' Parser function exposed to the configurator modules.

        :param resource: Resource name (HEALTHMONITOR/INTERFACES/ROUTES/
        FIREWALL/LOADBALANCER/LOADBALANCERV2/VPN)
        :param data: Resource data dictionary in case of device configuration
        and context in case of service configuration

        Returns: a dictionary if nfds/nfs contains a single element else
                 a list of dictionaries where each dictionary corresponds
                 to each element in nfds/nfs
        '''

        config_data_list = []

        if data.get('nfds'):
            tenant_id = data['tenant_id']
            nf_config_list = data['nfds']
        elif data.get('resource_data'):
            tenant_id = data['resource_data']['tenant_id']
            nf_config_list = data['resource_data']['nfs']
        else:
            msg = ("The given schema of data dictionary is not supported "
                   "by the data parser library. Returning the input. "
                   "Input data is: %s" % data)
            LOG.debug(msg)
            return data

        for nf_config in nf_config_list:
            self.resource_data = {}
            self.resource_data.update({
                'tenant_id': tenant_id,
                'role': nf_config['role'],
                'mgmt_ip': nf_config['svc_mgmt_fixed_ip']})

            self._parse_config_data(nf_config, resource)
            config_data_list.append(copy.deepcopy(self.resource_data))

        return (config_data_list[0]
                if len(config_data_list) == 1
                else config_data_list)

    def _parse_config_data(self, nfd, resource):
        if resource.lower() == const.HEALTHMONITOR_RESOURCE:
            return self.resource_data.update(
                {'periodicity': nfd['periodicity'],
                 'vmid': nfd['vmid']})

        networks = nfd['networks']
        for network in networks:
            prefix = network['type']
            port = network['ports'][0]
            self.resource_data.update({
                (prefix + '_cidr'): network['cidr'],
                (prefix + '_ip'): port['fixed_ip'],
                (prefix + '_floating_ip'): port['floating_ip'],
                (prefix + '_mac'): port['mac'],
                (prefix + '_gw_ip'): network['gw_ip']})

        vips = nfd.get('vips')
        if not vips:
            return
        for vip in vips:
            prefix = vip['type'] + '_vip'
            self.resource_data.update({
                (prefix + '_ip'): vip['ip'],
                (prefix + '_mac'): vip['mac']})



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import commands
import logging
import sys

from libs import config_libs
from libs import utils_libs
from libs import verify_libs


def main():

    # Run the Testcase:
    test = test_gbp_ri_func_2()
    test.run()


class test_gbp_ri_func_2(object):

    # Initialize logging
    logging.basicConfig(
        format='%(asctime)s [%(levelname)s] %(name)s - %(message)s',
        level=logging.WARNING)
    _log = logging.getLogger(__name__)
    cmd = 'rm /tmp/test_gbp_ri_func_2.log'
    commands.getoutput(cmd)
    hdlr = logging.FileHandler('/tmp/test_gbp_ri_func_2.log')
    formatter = logging.Formatter('%(asctime)s %(levelname)s %(message)s')
    hdlr.setFormatter(formatter)
    _log.addHandler(hdlr)
    _log.setLevel(logging.INFO)
    _log.setLevel(logging.DEBUG)

    def __init__(self):
        """
        Init def
        """
        self.gbpcfg = config_libs.Gbp_Config()
        self.gbpverify = verify_libs.Gbp_Verify()
        self.l3pol_name = 'demo_l3pol'
        self.l2pol_name = 'demo_l2pol'
        self.ptg_name = 'demo_ptg'

    def cleanup(self, cfgobj, uuid_name, tc_name='', fail=0):
        if isinstance(cfgobj, str):
            cfgobj = [cfgobj]
        if isinstance(uuid_name, str):
            uuid_name = [uuid_name]
        for obj, _id in zip(cfgobj, uuid_name):
            if self.gbpcfg.gbp_policy_cfg_all(0, obj, _id):
                self._log.info(
                    'Success in Clean-up/Delete of Policy Object %s\n' %
                    (obj))
            else:
                self._log.info(
                    'Failed to Clean-up/Delete of Policy Object %s\n' %
                    (obj))
        if fail != 0:
            self._log.info("\n## TESTCASE_GBP_RI_FUNC_2: FAILED")
            commands.report_results('test_gbp_ri_func_2', 'test_results.txt')
            sys.exit(1)

    def run(self):
        self._log.info(
            "\n## TESTCASE_GBP_RI_FUNC_2A: RESOURCE INTEGRITY AMONG "
            "L2POLICY and L3POLICY OBJs")
        # Testcase work-flow starts
        # ============ ALL POLICY OBJECTS ARE TO BE CREATED AND VERIFIED =
        self._log.info("\n##  Step 1: Create L3Policy ##\n")
        l3p_uuid = self.gbpcfg.gbp_policy_cfg_all(1, 'l3p', self.l3pol_name)
        if l3p_uuid == 0:
            self._log.info("# Step 1: Create L3Policy == Failed")
            self.cleanup(
                'l3p',
                l3p_uuid,
                tc_name='TESTCASE_GBP_RI_FUNC_2A',
                fail=1)
        ######
        self._log.info("\n## Step 2: Create L2Policy using L3Policy ##\n")
        l2p_uuid = self.gbpcfg.gbp_policy_cfg_all(
            1, 'l2p', self.l2pol_name, l3_policy=l3p_uuid)
        # this is needed for cleanup,can append and sort for the sake
        # of order.. but it kept it simple. l2p_uuid[1] is same as l3p_uuid,
        # just that cfg_all func returns both uuid when l2p is obj
        objs, names = ['l2p', 'l3p'], [l2p_uuid[0], l2p_uuid[1]]
        if l2p_uuid == 0:
            self._log.info("# Step 2: Create L2Policy == Failed")
            self.cleanup(objs, names, fail=1)
        self._log.info("\n## Step 3: Delete in-use L3Policy ##\n")
        if self.gbpcfg.gbp_policy_cfg_all(0, 'l3p', l3p_uuid) != 0:
            self._log.info(
                "\n# Step 4A: Delete in-use L3Policy did not fail #")
            self.cleanup(
                objs,
                names,
                tc_name='TESTCASE_GBP_RI_FUNC_2A',
                fail=1)
        else:
            self._log.info("\n## TESTCASE_GBP_RI_FUNC_2A: PASSED")
        ######
        self._log.info(
            "\n## TESTCASE_GBP_RI_FUNC_2B: RESOURCE INTEGRITY AMONG L2POLICY "
            "and PTG OBJs")
        self._log.info("\n## Step 5: Create Policy Target-Grp ##\n")
        uuids = self.gbpcfg.gbp_policy_cfg_all(
            1, 'group', self.ptg_name, l2_policy=l2p_uuid[0])
        if uuids != 0:
            objs, names = ['group', 'l2p', 'l3p'],\
                [self.ptg_name, l2p_uuid[0], l3p_uuid]
        else:
            self._log.info("# Step 5: Create Policy Target-Grp == Failed")
            self.cleanup(
                objs,
                names,
                tc_name='TESTCASE_GBP_RI_FUNC_2B',
                fail=1)

        self._log.info("\n## Step 5A: Delete in-use L2 Policy ##")
        if self.gbpcfg.gbp_policy_cfg_all(0, 'l2p', l2p_uuid[0]) != 0:
            self._log.info("\n# Step 5A: Delete in-use L2Policy did not fail")
            self.cleanup(
                objs,
                names,
                tc_name='TESTCASE_GBP_RI_FUNC_2B',
                fail=1)
        else:
            self._log.info("\n## TESTCASE_GBP_RI_FUNC_2B: PASSED")
        self.cleanup(objs, names)  # Cleanup the system now
        utils_libs.report_results('test_gbp_ri_func_2', 'test_results.txt')
        sys.exit(1)

if __name__ == '__main__':
    main()
